package it.unimi.dsi.webgraph;

/*		 
 * Copyright (C) 2003, 2004, 2005 Paolo Boldi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StreamTokenizer;
import java.lang.reflect.InvocationTargetException;
import java.util.NoSuchElementException;


/** An extension of {@link ImmutableGraph} that corresponds to graphs stored in a human-readable
 *  ASCII format. The file format is as follows: the graph is stored in a file named <samp>BASENAME+".graph-txt"</samp>.
 *  The first line contains the number of nodes, <var>n</var>. Then, <var>n</var> lines follow, the <var>i</var>-th
 *  line representing the successors of node <var>i</var> (nodes are numbered from 0 to <var>n</var>&minus;1).
 *  Successors are separated by a single space.
 *
 *  <P>This class is intended for testing and debugging purposes: it is neither memory nor time efficient.
 */


public class ASCIIGraph extends ImmutableGraph {

	/** Number of nodes. */
	private int n;
	/** Number of arcs. */
	private long m;
	/** Vector of successors. */
	private IntArrayList succ[];

	public int numNodes() {
		return n;
	}

	public long numArcs() {
		return m;
	}

	public IntIterator successors( final int x ) {
		return new AbstractIntIterator() {
				private IntArrayList l = succ[ x ];
				private int outd = l.size();
				private int p = 0;
				public int nextInt() {
					if ( p < outd ) return l.getInt( p++ );
					else throw new NoSuchElementException();
				}
				public Object next() { return new Integer( nextInt() ); }
				public boolean hasNext() { return ( p < outd ); }
				public void remove() { throw new UnsupportedOperationException(); }
			};
	}

	public int outdegree( int x ) {
		return succ[ x ].size();
	}

	public static ImmutableGraph loadSequential( CharSequence baseName ) throws IOException {
		return load( baseName );
	}

	public static ImmutableGraph loadSequential( CharSequence baseName, ProgressMeter unused ) throws IOException {
		if ( false && unused != null );
		return load( baseName, null );
	}

	public static ImmutableGraph loadOffline( CharSequence baseName ) throws IOException {
		return load( baseName, null );
	}

	public static ImmutableGraph loadOffline( CharSequence baseName, ProgressMeter unused ) throws IOException {
		if ( false && unused != null );
		return load( baseName, null );
	}

	public static ImmutableGraph load( CharSequence baseName ) throws IOException {
		return load( baseName, null );
	}

	public static ImmutableGraph load( CharSequence baseName, ProgressMeter unused ) throws IOException {
		if ( false && unused != null );
		ASCIIGraph result = new ASCIIGraph();
		StreamTokenizer st = new StreamTokenizer( new BufferedReader( new FileReader( baseName + ".graph-txt" ) ) );
		st.eolIsSignificant( true );
		st.parseNumbers();

		int tokenType, n, dep;

		if ( st.nextToken() != StreamTokenizer.TT_NUMBER ) // Read number of nodes
			throw new IOException( "Number of nodes expected at line 1" ); 
		n = result.n = (int)st.nval;
		if ( n < 0 ) throw new IOException( "Number of nodes must be nonnegative" );
		result.succ = new IntArrayList[ result.n ];
		if ( st.nextToken() != StreamTokenizer.TT_EOL ) // Ignore newline
			throw new IOException( "Number of nodes must be followed by a newline at line 1" );

		for ( int i = 0; i < n; i++ ) {
			result.succ[ i ] = new IntArrayList();
			do {
				tokenType = st.nextToken();
				if ( tokenType == StreamTokenizer.TT_NUMBER ) {
					result.succ[ i ].add( dep = (int)st.nval );
					if ( dep < 0 || dep >= n )
						throw new IOException( "The value " + dep + " is not a node index at line " + st.lineno() );
					result.m++;
				} else if ( tokenType != StreamTokenizer.TT_EOL )
					throw new IOException( "Unrecognized character sequence at line " + st.lineno() );
			} while ( tokenType != StreamTokenizer.TT_EOL );
		}

		
		return result;
	}

	public static void store( ImmutableGraph graph, CharSequence baseName, ProgressMeter unused ) throws IOException {
		if ( false ) if ( unused != null );
		store( graph, baseName );
	}

	public static void store( ImmutableGraph graph, CharSequence baseName ) throws IOException {
		PrintWriter pw = new PrintWriter( new BufferedWriter( new FileWriter( baseName + ".graph-txt" ) ) );
		int n = graph.numNodes();
		IntIterator ii;

		pw.println( n );
		for ( NodeIterator nodeIterator = graph.nodeIterator(); nodeIterator.hasNext(); ) {
			nodeIterator.next();
			nodeIterator.outdegree();
			ii = nodeIterator.successors();
			while ( ii.hasNext() ) pw.print( ii.nextInt() + " " );
			pw.println();
		}
		pw.close();
	}

	public static void main( String args[] ) throws IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, IOException, ClassNotFoundException  {
		String source = null, dest = null;
		Class graphClass = null;

		int ch;
		LongOpt longopts[] = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "graph-class", LongOpt.REQUIRED_ARGUMENT, null, 'g' ),
		};

		Getopt g = new Getopt( "ASCIIGraph", args, "c:w:m:g:h", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: ASCIIGraph [OPTIONS] SOURCE DEST" );
			System.err.println( "Reads a graph with basename SOURCE and writes it out" );
			System.err.println( "with basename DEST in ASCII format." );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -g, --graph-class   forces a Java class to load the source graph" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help          print this help screen" );
			System.err.println( "" );
			return;

		case '?':
			System.err.println( "ASCIIGraph: unknown option " + ( char ) g.getOptopt() );
			System.err.println( "Try 'ASCIIGraph --help' for more information." );
			return;

		case 'g':
			graphClass = Class.forName( g.getOptarg(), true, ClassLoader.getSystemClassLoader() );
			break;

		default:
			break;
		}

		if ( args.length - g.getOptind() < 2 ) {
			System.err.println( "Wrong number (" + ( args.length - g.getOptind() ) + ") of arguments." );
			return;
		}
		else {
			source = args[ g.getOptind() ];
			dest = args[ g.getOptind() + 1 ];
		}

		final ProgressMeter pm = new ProgressMeter();

		final ImmutableGraph graph = graphClass != null 
			? (ImmutableGraph)graphClass.getMethod( "loadSequential", new Class[] { CharSequence.class, ProgressMeter.class } ).invoke( graphClass, new Object[] { source, pm } )
			: ImmutableGraph.loadSequential( source, pm );
		ASCIIGraph.store( graph, dest );
	}
}




// Local Variables:
// mode: jde
// tab-width: 4
// End:
