package it.unimi.dsi.webgraph;

/*		 
 * Copyright (C) 2003, 2004, 2005 Paolo Boldi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

// TODO check what we do if the node number is out of range

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.Iterators;
import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntIterators;
import it.unimi.dsi.mg4j.index.CompressionFlags;
import it.unimi.dsi.mg4j.io.FastByteArrayOutputStream;
import it.unimi.dsi.mg4j.io.FastMultiByteArrayInputStream;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.NullOutputStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.InvocationTargetException;
import java.util.NoSuchElementException;
import java.util.Properties;


/** An immutable graph represented using the techniques described in 
 * <a href="http://vigna.dsi.unimi.it/papers.php#BoVWFI"><i>The WebGraph Framework I: Compression Techniques</i></a>, by Paolo Boldi and
 * Sebastiano Vigna, in Proc&#46; of the Thirteenth World&ndash;Wide Web
 * Conference, pages 595&minus;601, 2004, ACM Press.
 *
 * <P>This class provides a flexible and configurable way to storing and
 * accessing web graphs in a compressed form.  Its main method can load an
 * {@link ImmutableGraph} and compress it. The resulting compressed {@link
 * BVGraph} is described by a <em>graph file</em> (with extension
 * <samp>.graph</samp>), an <em>offset file</em> (with extension
 * <samp>.offsets</samp>) and a <em>property file</em> (with extension
 * <samp>.properties</samp>). The latter, not surprisingly, is a Java property file.
 *
 *
 * <h2>The Graph File</h2>
 *
 * <P>This class stores a graph as an <a
 * href="http://mg4j.dsi.unimi.it/">MG4J</A> bit stream. The bit stream format
 * depends on a number of parameters and encodings that can be mixed
 * orthogonally. The parameters are:
 * 
 * <ul>
 *
 * <li>the <em>window size</em>, a nonnegative integer;
 * <li>the <em>maximum reference count</em>, a positive integer (it is meaningful only when the window is nonzero);
 * <li>the <em>minimum interval length</em>, an integer larger than or equal to two, or 0, which is interpreted as infinity.
 *
 * </ul>
 *
 * <H3>Successor Lists</H3>
 * 
 * <P>The graph file is a sequence of successor lists, one for each node. 
 * The list of node <var>x</var> can be thought of as a sequence of natural numbers (even though, as we will
 * explain later, this sequence is further coded suitably as a sequence of bits):
 * <OL STYLE="list-style-type: lower-alpha">
 *  <LI>The outdegree of the node; if it is zero, the list ends here.
 *  <LI>If the window size is not zero, the <em>reference part</em>, that is:
 *    <OL><LI>a nonnegative integer, the <em>reference</em>, which never exceeds the window size; if the reference
 *       is <var>r</var>, the list of successors will be specified as a modified version of the list of successors
 *       of <var>x</var>&minus;<var>r</var>; if <var>r</var> is 0, then the list of successors will be specified 
 *       explicitly;
 *      <LI>if <var>r</var> is nonzero:
 *         <OL STYLE="list-style-type: lower-roman">
 *            <LI>a natural number <var>b</var>, the <em>block count</em>;
 *            <LI>a sequence of <var>b</var> natural numbers <var>B</var><sub>1</sub>, &hellip;, <var>B</var><sub>b</sub>, called the <em>copy-block list</em>; only the
 *            first number can be zero.
 *        </OL>
 *
 *    </OL>
 *  <LI>Then comes the <em>extra part</em>, specifying some more entries that the list of successors contains (or all of them, if
 *   <var>r</var> is zero), that is:
 *    <OL>
 *      <LI>If the minimum interval length is finite,
 *       <OL STYLE="list-style-type: lower-roman">
 *          <LI>an integer <var>i</var>, the <em>interval count</em>;
 *          <LI>a sequence of <var>i</var> pairs, whose first component is the left extreme of an interval,
 *              and whose second component is the length of the interval (i.e., the number of integers contained in the interval).
 *       </OL>
 *      <li>Finally, the list of <em>residuals</em>, which contain all successors not specified by previous methods.
 *    </OL>
 *  </OL>
 *       
 *    <P>The above data should be interpreted as follows: 
 *    <ul>
 *      <li>The reference part, if present (i.e., if both the window size and the reference are strictly positive), specifies
 *          that part of the list of successors of node <var>x</var>&minus;<var>r</var> should be copied; the successors of
 *          node <var>x</var>&minus;<var>r</var> that should be copied are described in the copy-block list; more precisely, one should copy
 *          the first <var>B</var><sub>1</sub> entries of this list, discard the next <var>B</var><sub>2</sub>, copy
 *          the next <var>B</var><sub>3</sub> etc. (the last remaining elements of the list of successors will be copied if <var>b</var> is
 *          even, and discarded if <var>b</var> is odd).
 *      <li>The extra part specifies additional successors (or all of them, if the reference part is absent); the extra part is not present
 *          if the number of successors that are to be copied according to the reference part already coincides with the outdegree of <var>x</var>;
 *          the successors listed in the extra part are given in two forms:
 *          <ul>
 *            <li>some of them are specified as belonging to (integer) intervals, if the minimum interval length is finite; 
 *              the interval count indicates how many intervals,
 *              and the intervals themselves are listed as pairs (left extreme, length);
 *            <li>the residuals are the remaining "scattered" successors.
 *          </ul>
 *    </ul>
 *
 * 
 * <H3>How Successor Lists Are Coded</H3>
 *
 * <P>As we said before, the list of integers corresponding to each successor list should be coded into a sequence of bits.
 * This is (ideally) done in two phases: we first modify the sequence in a suitable manner (as explained below) so to obtain
 * another sequence of integers (some of them might be negative). Then each single integer is coded, using a coding that can
 * be specified as an option; the integers that may be negative are first turned into natural numbers using {@link Fast#int2nat(int)}. 
 *
 * <OL>
 *  <LI>The outdegree of the node is left unchanged, as well as the reference and the block count;
 *  <LI>all blocks are decremented by 1, except for the first one;
 *  <LI>the interval count is left unchanged;
 *  <LI>all interval lengths are decremented by the minimum interval length;
 *  <LI>the first left extreme is expressed as its difference from <var>x</var> (it will be negative if the first extreme is
 *     less than <var>x</var>); the remaining left extremes are expressed as their distance from the previous right extreme
 *     plus 2 (e.g., if the interval is [5..11] and the previous one was [1..3], then the left extreme 5 is expressed as
 *     5-(3+2)=5-5=0);
 *  <LI>the first residual is expressed as its difference from <var>x</var> (it will be negative if the first residual is
 *     less than <var>x</var>); the remaining residuals are expressed as decremented differences from the previous residual.
 *  </OL>
 *
 * <H2>The Offset File</H2>
 *
 * <P>Since the graph is stored as a bit stream, we must have some way to know where each successor list starts. 
 * This information is stored in the offset file, which contains the bit offset of each successor list (in particular,
 * the offset of the first successor list will be zero). As a commodity, the offset file contains an additional
 * offset pointing just after the last successor list (providing, as a side-effect, the actual bit length of the graph file).
 * Each offset (except for the first one) is stored as a suitably coded difference from the previous offset.
 *
 * <H2>The Property File</H2>
 * 
 * <P>This file contains self-explaining entries that are necessary to correctly decode the graph and offset files, and
 * moreover give some statistical information about the compressed graph (e.g., the number of bits per link).
 * 
 *
 * <H2>How The Graph File Is Loaded Into Memory</H2>
 *
 * <P>The natural way of using a graph file is to load it into a byte array and
 * then index its bits using the suitable offset. This class will use a byte
 * array for graphs smaller than {@link Integer#MAX_VALUE} bytes,
 * and a {@link it.unimi.dsi.mg4j.io.FastMultiByteArrayInputStream}
 * otherwise: in the latter case, expect a significant slowdown (as
 * an {@link it.unimi.dsi.mg4j.io.InputBitStream} can wrap directly
 * a byte array). 
 * 
 * <P>Unfortunately, offsets are
 * very memory expensive. For this reason, this class provides as an option to
 * load offsets only partially; more precisely, one can specify an <em>offset
 * step</em> <var>J</var>, and load just one offset each <var>J</var>. Since a
 * successor list can be in general skipped only knowing the successors of the
 * referenced list (as we have otherwise no way of knowing the length of the
 * residuals), this would require a possibly large backward recursion. Thus,
 * this class can load a graph in a special rearranged form (some of the bits
 * are permuted) that makes it possible to recover each successor list
 * independently: for each <var>J</var> successors lists we first store
 * the outdegrees, and then the rest of each successor list. This is called a
 * <em>successor-list block</em> (of course, if <var>J</var> is one than a
 * successor-list block is exactly a successor list).  
 * <P>Optionally, this class may load no offets at all. In this case, the only
 * way to access the graph is by creating a {@link #nodeIterator()}.
 *
 * <H2>Not Loading the Graph File at All</H2>
 * 
 * <P>For some applications (such as transposing a graph) it is not necessary to load the graph
 * file in memory. Since this class is able to enumerate the links of a graph without using random
 * access, it is possible not to load in memory any information at all, and obtain iterators that
 * directly read from the graph file. To obtain this effect, you must call {@link #loadOffline(CharSequence)}.
 *
 **/


public class BVGraph extends ImmutableGraph implements CompressionFlags {

	private static final int STD_BUFFER_SIZE = 1024 * 1024;

	/** This number classifies the present graph format. When new features require introducing binary incompatibilities,
		this number is bumped so to ensure that old classes do not try to read graphs they cannot understand. */
	public final static int BVGRAPH_VERSION = 0;

	/** The initial length of an array that will contain a successor list. */
	protected static final int INITIAL_SUCCESSOR_LIST_LENGTH = 1024;
	
	/** A special value for {@link #minIntervalLength} interpreted as meaning that the minimum interval length is infinity. */
	protected static final int NO_INTERVALS = 0;
	/** The basename of the graph. This may be <code>null</code>, but trying to load the graph with an offset
	 * step of -1 will cause an exception. */
	protected CharSequence basename;
	
	/** The number of nodes of the graph. */
	protected int n;

	/** The number of arcs of the graph. */
	protected long m;

	/** When {@link #offsetStep} is not -1, whether this graph is directly loaded into 
	 * {@link #graphMemory}, or rather wrapped in a {@link it.unimi.dsi.mg4j.io.FastMultiByteArrayInputStream}
	 * specified by {@link #graphStream}. */
	protected boolean isMemory;

	/** The byte array storing the compressed graph, if {@link #isMemory} is true and {@link #offsetStep} is not -1.
	 *  
	 * <P>This variable is loaded with a copy of the graph file, or with
	 * a rearrangement of the latter, depending on whether {@link #offsetStep} is smaller than or equal to one. If
	 * {@link #offsetStep} is -1, this variable is <code>null</code>, and node iterators are generated by opening
	 * streams directly on the graph file. */
	protected byte graphMemory[];

	/** The long array input stream storing the compressed graph, if {@link #isMemory} is false and {@link #offsetStep} is not -1. 
	 * 
	 * <P>It is loaded with a copy of the graph file. If
	 * {@link #offsetStep} is -1, this variable is <code>null</code>, and node iterators are generated by opening
	 * streams directly on the graph file. */
	protected FastMultiByteArrayInputStream graphStream;

	/** This variable is <code>null</code> iff {@link #offsetStep} is zero or less
	 * (impliying that offsets have not been loaded).  Otherwise, the entry of
	 * index <var>i</var> represent the offset (in bits) at which each the
	 * <var>i</var>-th successor-list block starts. The last entry contains the
	 * length in bits of the whole graph file. */
	protected long offset[];

	/** The maximum reference count. */
	protected int maxRefCount = DEFAULT_MAX_REF_COUNT;

	/** Default backward reference maximum length (no bound). */
	public final static int DEFAULT_MAX_REF_COUNT = Integer.MAX_VALUE;

	/** The window size. Zero means no references. */
	protected int windowSize = DEFAULT_WINDOW_SIZE;

	/** Default window size. */
	public final static int DEFAULT_WINDOW_SIZE = 7;

	/** The minimum interval length. */
	protected int minIntervalLength = DEFAULT_MIN_INTERVAL_LENGTH;

	/** Default minimum interval length. */
	public final static int DEFAULT_MIN_INTERVAL_LENGTH = 3;

	/** The offset step. Special values are 0, meaning no offsets, and -1, meaning no offset and no 
	 * {@link #graphMemory}. 
	 */
	protected int offsetStep = DEFAULT_OFFSET_STEP;

	/** Default offset step. */
	public final static int DEFAULT_OFFSET_STEP = 1;

	/** The value of <var>k</var> for &zeta;<sub><var>k</var></sub> coding (for residuals). */
	protected int zetaK = DEFAULT_ZETA_K;

	/** Default value of <var>k</var>. */
	public final static int DEFAULT_ZETA_K = 3;

	/** Flag: write outdegrees using &gamma; coding (default). */
	public static final int OUTDEGREES_GAMMA = GAMMA;

	/** Flag: write outdegrees using &delta; coding. */
	public static final int OUTDEGREES_DELTA = DELTA;

	/** Flag: write copy-block lists using &gamma; coding (default). */
	public static final int BLOCKS_GAMMA = GAMMA << 4;

	/** Flag: write copy-block lists using &delta; coding. */
	public static final int BLOCKS_DELTA = DELTA << 4;

	/** Flag: write residuals using &gamma; coding. */
	public static final int RESIDUALS_GAMMA = GAMMA << 8;

	/** Flag: write residuals using &zeta;<sub><var>k</var></sub> coding (default). */
	public static final int RESIDUALS_ZETA = ZETA << 8;

	/** Flag: write residuals using &delta; coding. */
	public static final int RESIDUALS_DELTA = DELTA << 8;

	/** Flag: write residuals using variable-length nibble coding. */
	public static final int RESIDUALS_NIBBLE = NIBBLE << 8;

	/** Flag: write references using &gamma; coding. */
	public static final int REFERENCES_GAMMA = GAMMA << 12;

	/** Flag: write references using &delta; coding. */
	public static final int REFERENCES_DELTA = DELTA << 12;

	/** Flag: write references using unary coding (default). */
	public static final int REFERENCES_UNARY = UNARY << 12;

	/** Flag: write block counts using &gamma; coding (default). */
	public static final int BLOCK_COUNT_GAMMA = GAMMA << 16;

	/** Flag: write block counts using &delta; coding. */
	public static final int BLOCK_COUNT_DELTA = DELTA << 16;

	/** Flag: write block counts using unary coding. */
	public static final int BLOCK_COUNT_UNARY = UNARY << 16;

	/** Flag: write offsets using &gamma; coding (default). */
	public static final int OFFSETS_GAMMA = GAMMA << 20;

	/** Flag: write offsets using &delta; coding. */
	public static final int OFFSETS_DELTA = DELTA << 20;

	/** The coding for outdegrees. By default, we use &gamma; coding. */
	protected int outdegreeCoding = GAMMA;

	/** The coding for copy-block lists. By default, we use &gamma; coding. */
	protected int blockCoding = GAMMA;

	/** The coding for residuals. By default, we use &zeta; coding. */
	protected int residualCoding = ZETA;

	/** The coding for references. By default, we use unary coding. */
	protected int referenceCoding = UNARY;

	/** The coding for block counts. By default, we use &gamma; coding. */
	protected int blockCountCoding = GAMMA;

	/** The coding for offsets. By default, we use &gamma; coding. */
	protected int offsetCoding = GAMMA;

	/** The compression flags used. */
	private int flags = 0;

	private final static boolean STATS = false;
	private final static boolean DEBUG = false;
	private final static boolean ASSERTS = false;

	static { if ( DEBUG && ASSERTS && STATS ); }

	private PrintWriter offsetStats, outdegreeStats, blockCountStats, blockStats, intervalCountStats, referenceStats, leftStats, lenStats, residualStats, residualCountStats;

	protected BVGraph() {}

	public int numNodes() {
		return n;
	}

	public long numArcs() {
		return m;
	}
	
	public CharSequence basename() {
		return basename;
	}

	/** Returns the maximum reference count of this graph. 
	 *
	 * @return the maximum reference count.
	 */
	public int maxRefCount() {
		return maxRefCount;
	}

	/** Returns the window size of this graph. 
	 *
	 * @return the window size.
	 */
	public int windowSize() {
		return windowSize;
	}

	/** Returns the offset step of this graph. 
	 *
	 * @return the offset step.
	 */
	public int offsetStep() {
		return offsetStep;
	}


	/* This family of protected methods is used throughout the class to read data
	from the graph file following the codings indicated by the compression
	flags. */

	/** Reads an offset difference from the given stream. 
	 *
	 * @param ibs an offset-file input bit stream.
	 * @return the next offset difference.
	 */
	protected int readOffset( InputBitStream ibs ) throws IOException {
		switch( offsetCoding ) {
		case GAMMA: return ibs.readGamma(); 
		case DELTA: return ibs.readDelta(); 
		default: throw new UnsupportedOperationException( "The required offset coding (" + offsetCoding + ") is not supported." );
		}
	}

	/** Writes an offset difference to the given stream. 
	 *
	 * @param obs an offset-file output bit stream.
	 * @param x an offset difference to be stored in the stream.
	 * @return the number of bits written.
	 */
	protected int writeOffset( OutputBitStream obs, int x ) throws IOException {
		switch( offsetCoding ) {
		case GAMMA: return obs.writeGamma( x ); 
		case DELTA: return obs.writeDelta( x ); 
		default: throw new UnsupportedOperationException( "The required offset coding (" + offsetCoding + ") is not supported." );
		}
	}

	/** Reads an outdegree from the given stream. 
	 *
	 * @param ibs a graph-file input bit stream.
	 * @return the next outdegree.
	 */
	protected int readOutdegree( InputBitStream ibs ) throws IOException {
		switch( outdegreeCoding ) {
		case GAMMA: return ibs.readGamma(); 
		case DELTA: return ibs.readDelta(); 
		default: throw new UnsupportedOperationException( "The required outdegree coding (" + outdegreeCoding + ") is not supported." );
		}
	}

	/** Writes an outdegree to the given stream. 
	 *
	 * @param obs a graph-file output bit stream.
	 * @param d an outdegree to be stored in the stream.
	 * @return the number of bits written.
	 */
	protected int writeOutdegree( OutputBitStream obs, int d ) throws IOException {
		switch( outdegreeCoding ) {
		case GAMMA: return obs.writeGamma( d ); 
		case DELTA: return obs.writeDelta( d ); 
		default: throw new UnsupportedOperationException( "The required outdegree coding (" + outdegreeCoding + ") is not supported." );
		}
	}

	/** Reads a reference from the given stream. 
	 *
	 * @param ibs a graph-file input bit stream.
	 * @return the next reference.
	 */
	protected int readReference( InputBitStream ibs ) throws IOException {
		final int ref;

		switch( referenceCoding ) {
		case UNARY: ref = ibs.readUnary(); break;
		case GAMMA: ref = ibs.readGamma(); break;
		case DELTA: ref = ibs.readDelta(); break;
		default: throw new UnsupportedOperationException( "The required reference coding (" + referenceCoding + ") is not supported." );
		}
		if ( ref > windowSize ) throw new IllegalStateException( "The required reference (" + ref + ") is incompatible with the window size (" + windowSize + ")" );
		return ref;
	}
	
	/** Writes a reference to the given stream. 
	 *
	 * @param obs a graph-file output bit stream.
	 * @param ref the reference.
	 * @return the number of bits written.
	 */
	protected int writeReference( OutputBitStream obs, int ref ) throws IOException {

		if ( ref > windowSize ) throw new IllegalStateException( "The required reference (" + ref + ") is incompatible with the window size (" + windowSize + ")" );
		switch( referenceCoding ) {
		case UNARY: return obs.writeUnary( ref );
		case GAMMA: return obs.writeGamma( ref );
		case DELTA: return obs.writeDelta( ref );
		default: throw new UnsupportedOperationException( "The required reference coding (" + referenceCoding + ") is not supported." );
		}
	}


	/** Reads a block count from the given stream. 
	 *
	 * @param ibs a graph-file input bit stream.
	 * @return the next block count.
	 */
	protected int readBlockCount( InputBitStream ibs ) throws IOException {
		switch( blockCountCoding ) {
		case UNARY: return ibs.readUnary();
		case GAMMA: return ibs.readGamma();
		case DELTA: return ibs.readDelta();
		default: throw new UnsupportedOperationException( "The required block count coding (" + blockCountCoding + ") is not supported." );
		}
	}

	/** Writes a block count to the given stream. 
	 *
	 * @param obs a graph-file output bit stream.
	 * @param count the block count.
	 * @return the number of written bits.
	 */
	protected int writeBlockCount( OutputBitStream obs, int count ) throws IOException {
		switch( blockCountCoding ) {
		case UNARY: return obs.writeUnary( count );
		case GAMMA: return obs.writeGamma( count );
		case DELTA: return obs.writeDelta( count );
		default: throw new UnsupportedOperationException( "The required block count coding (" + blockCountCoding + ") is not supported." );
		}
	}


	/** Reads a block from the given stream. 
	 *
	 * @param ibs a graph-file input bit stream.
	 * @return the next block.
	 */
	protected int readBlock( InputBitStream ibs ) throws IOException {
		switch( blockCoding ) {
		case UNARY: return ibs.readUnary();
		case GAMMA: return ibs.readGamma();
		case DELTA: return ibs.readDelta();
		default: throw new UnsupportedOperationException( "The required block coding (" + blockCoding + ") is not supported." );
		}
	}

	/** Writes a block to the given stream. 
	 *
	 * @param obs a graph-file output bit stream.
	 * @param block the block.
	 * @return the number of written bits.
	 */
	protected int writeBlock( OutputBitStream obs, int block ) throws IOException {
		switch( blockCoding ) {
		case UNARY: return obs.writeUnary( block );
		case GAMMA: return obs.writeGamma( block );
		case DELTA: return obs.writeDelta( block );
		default: throw new UnsupportedOperationException( "The required block coding (" + blockCoding + ") is not supported." );
		}
	}
	
	/** Reads a residual from the given stream. 
	 *
	 * @param ibs a graph-file input bit stream.
	 * @return the next residual.
	 */
	protected int readResidual( InputBitStream ibs ) throws IOException {
		switch( residualCoding ) {
		case GAMMA: return ibs.readGamma();
		case ZETA: return ibs.readZeta( zetaK );
		case DELTA: return ibs.readDelta();
		case NIBBLE: return ibs.readNibble();
		default: throw new UnsupportedOperationException( "The required residuals coding (" + residualCoding + ") is not supported." );
		}
	}

	/** Writes a residual to the given stream. 
	 *
	 * @param obs a graph-file output bit stream.
	 * @param residual the residual.
	 * @return the number of written bits.
	 */
	protected int writeResidual( OutputBitStream obs, int residual ) throws IOException {
		switch( residualCoding ) {
		case GAMMA: return obs.writeGamma( residual );
		case ZETA: return obs.writeZeta( residual, zetaK );
		case DELTA: return obs.writeDelta( residual );
		case NIBBLE: return obs.writeNibble( residual );
		default: throw new UnsupportedOperationException( "The required residuals coding (" + residualCoding + ") is not supported." );
		}
	}

	/* We now define the core methods that access the graph stored in
	graphMemory or graphStream. There are a number of delicate issues here, as these method
	must be able to compute outdegree and successor lists positions even if
	offsetStep is greater than one.  Unfortunately, this requires a small amount of logic
	duplication in skipNode() and in successors(). */


	/** A bit stream wrapping {@link #graphMemory}, or {@link #graphStream}, used <em>only</em> by
	{@link #outdegree(int)}. It is declared here for efficiency reasons. */
	protected InputBitStream outdegreeIbs;

	/** A cache maintaining the outdegrees from {@link #cacheStart} (inclusive) to {@link #outdegreeCacheEnd} (exclusive). */
	protected int outdCache[];
	/** The first node in the outdegree cache. */
	protected int cacheStart = Integer.MAX_VALUE;
	/** The number of last node of the outdegree cache plus one. */
	protected int outdegreeCacheEnd = Integer.MAX_VALUE;

	/** A cache maintaining the offsets from {@link #cacheStart} (inclusive) to {@link #offsetCacheEnd} (exclusive). */
	protected long offsetCache[];
	/** The number of last node of the offset cache plus one. This is never greater than {@link #outdegreeCacheEnd}. */
	protected int offsetCacheEnd = Integer.MAX_VALUE;
	


	public int outdegree( int x ) throws IllegalStateException {

		if (x < 0 || x >= n) throw new IllegalArgumentException( "Node index out of range:" + x );

		/* Computing the outdegree is a most basic operation. Thus, it must be always
		   possible to compute the outdegree of a node independently of any other state
		   in a BVGraph. To this purpose, we have special-purpose input bit stream that
		   is used just to read outdegrees. */

		try {
			// Without offsets, we just give up.
			if ( offsetStep <= 0 ) throw new IllegalStateException( "You cannot compute the outdegree of a random node without offsets" );
			// With all offsets, we just position and read.
			if ( offsetStep == 1 ) {
				outdegreeIbs.position( offset[ x ] );
				return readOutdegree( outdegreeIbs );
			}

			// Otherwise, it could happen that the required outdegree is in the outdegree cache. 			
			if ( x >= cacheStart && x < outdegreeCacheEnd ) return outdCache[ x - cacheStart ];

			outdegreeIbs.position( offset[ x / offsetStep ] );
			// We now skip x % offsetStep outdegrees to get the right one.
			int i = x % offsetStep;
			while( i-- != 0 ) readOutdegree( outdegreeIbs );
			return readOutdegree( outdegreeIbs );
		}
		catch( IOException cantHappen ) { throw new RuntimeException( cantHappen ); }
	}


	/** Skips the part of the successor list of a node that comes after the outdegree. 
	 *
	 * <P>This method must be called with <code>ibs</code> positioned exactly at the
	 * beginning of the successor list of node <code>x</code>, excluding the outdegree,
	 * which must be provided in <code>outd</code>.
	 * 
	 * <P><strong>Warning</strong>: This method duplicates unavoidably part of the
	 * logic of {@link #successors(int, InputBitStream, int[][], int[], int[])}; the
	 * two methods must remain tightly coupled.
	 * 
	 * @param ibs the input bit stream, positioned exactly at the start of the successor list.
	 * @param x the node whose successor list is to be skipped.
	 * @param outd the outdegree of <code>x</code>.
	 * @throws IllegalStateException if called without offsets.
	 */
	protected void skipNode( InputBitStream ibs, int x, int outd ) throws IllegalStateException {
		final int ref, blockCount;
		int i, l, extraCount;

		if ( ASSERTS ) assert x >= 0;

		// Without offsets, we just give up.
		if ( offsetStep <= 0 ) throw new IllegalStateException( "You cannot skip a successor list without offsets" );

		if ( outd == 0 ) return; // No successors, we're over.

		try {
			// First of all, we check the reference if necessary.
			if ( windowSize > 0 ) ref = readReference( ibs );
			else ref = -1;

			if ( ref > 0 ) {
				blockCount = readBlockCount( ibs );

				int copied = 0, total = 0;

				for( i = 0; i < blockCount; i++ ) {
					l = readBlock( ibs ) + ( i == 0 ? 0 : 1 );
					total += l;
					if ( i % 2 == 0 ) copied += l;
				}

				if ( blockCount % 2 == 0 ) copied += outdegree( x - ref ) - total;
				extraCount = outd - copied;
			}
			else extraCount = outd;

			int intervalCount = 0; // Number of intervals
		
			if ( extraCount != 0 ) {
				if ( minIntervalLength != NO_INTERVALS && ( intervalCount = ibs.readGamma() ) != 0 ) {
					for ( i = 0; i < intervalCount; i++ ) {
						ibs.readGamma(); // Throw away left point.
						l = ibs.readGamma() + minIntervalLength;
						extraCount -= l;
					}
				}
			}
			
			while( extraCount-- != 0 ) readResidual( ibs );
		}
		catch( IOException cantHappen ) { throw new RuntimeException( cantHappen ); }
	}
	
	/** Positions the given input bit stream exactly before the successor list of the
	 * given node, just after the outdegree, which is returned.  
	 *
	 * <P>This method will modify the offset and the outdegree caches.
	 *  
	 * @param ibs an input bit stream wrapping a graph file.
	 * @param x a node.
	 * @throws IllegalStateException if called without offsets.
	 * @return the outdegree of <code>x</code>
	 */
	protected int position( InputBitStream ibs, int x ) {
		try {
			if ( ASSERTS ) assert x >= 0;
			
			// Without offsets, we just give up.
			if ( offsetStep <= 0 ) throw new IllegalStateException( "You cannot position a stream without offsets" );
		
			if ( offsetStep == 1 ) {
				ibs.position( offset[ x ] );
				return readOutdegree( ibs );
			}

			// If we happen to be inside the offset cache, we're done.
			if ( x >= cacheStart && x < offsetCacheEnd ) {
				ibs.position( offsetCache[ x - cacheStart ] );
				return outdCache[ x - cacheStart ];
			}

			final long blockStartOffset = offset[ x / offsetStep ];
			ibs.position( blockStartOffset );
			ibs.readBits( blockStartOffset ); // We set the number of read bits so to read it later to fill the offset cache.
			
			final int offsetInBlock = x % offsetStep;
			cacheStart = x - offsetInBlock;
			final int actualStep = Math.min( offsetStep, n - cacheStart );
			outdegreeCacheEnd = cacheStart + actualStep;

			// First, we skip outdegrees so to get to the successor lists.
			int i;
			for( i = 0; i < actualStep; i++ ) outdCache[ i ] = readOutdegree( ibs );

			// Then, we skip the lists before the one we want.
			for( i = 0; i < offsetInBlock; i++ ) {
				offsetCache[ i ] = ibs.readBits();
				skipNode( ibs, cacheStart + i, outdCache[ i ] );
			}
			offsetCache[ offsetInBlock ] = ibs.readBits();
			offsetCacheEnd = cacheStart + offsetInBlock + 1;

			return outdCache[ offsetInBlock ];
		}
		catch( IOException cantHappen ) { throw new RuntimeException( cantHappen ); }
	}

	/** Returns an iterator over the successors of a given node.
	 * 
	 * @param x a node.
	 * @return an iterator over the successors of the node.
	 */
	public IntIterator successors( final int x ) {
		// We just call successors(int, InputBitStream, int[][], int[], int[]) with
		// a newly created input bit stream and null elsewhere.
		if ( ASSERTS ) assert x >=0;

		if ( offsetStep <= 0 ) throw new UnsupportedOperationException( "Random access to successor lists is not possible with sequential or offline graphs" );

		final InputBitStream ibs = isMemory ? new InputBitStream( graphMemory ) : new InputBitStream( new FastMultiByteArrayInputStream( graphStream ) );
		return successors( x, ibs, null, null, null );
	}


	/** Given an {@link InputBitStream} wrapping a graph file, returns an iterator over the
	 * successors of a given node <code>x</code>.
	 * 
	 * <P>This method can be used in two different ways:
	 * <OL><LI>by providing a node and an input bit stream wrapping a graph file, it is possible
	 * to access the successor list of the node (provided that offsets have been loaded);
	 * <LI>by providing additional data, which essentially are used to keep some state
	 * about the graph, it is possible to perform an efficient sequential visit of all
	 * successor lists (even when no offsets were loaded).
	 * </OL>
	 * 
	 * <P>This method may modify the offset and the outdegree caches if <code>window</code> is <code>null</code>.
	 *  	 
	 * 
	 * 
	 * @param x a node.
	 * @param ibs an input bit stream wrapping a graph file. After this method returns, the state of <code>ibs</code> is undefined:
	 *   however, after the iterator returned is exhausted, <code>ibs</code> will positioned just after the successor list of <code>x</code>.
	 * @param window either <code>null</code>, or a double array with the following meaning: <code>window[(x-i) mod offsetStep]</code> 
	 *   contains, for all <code>i</code> between 1 (inclusive) and {@link #windowSize} (exclusive), the list of successors
	 *   of node <code>x</code>&minus;<code>i</code>. If <code>window</code> is not <code>null</code> then <code>ibs</code>
	 *   must be positioned before the successor list of <code>x</code>; note that, if <code>offsetStep</code>
	 *   is greater than 1 and <code>x</code> is a multiple of <code>offsetStep</code>, we are positioned at the
	 *   start of a block (meaning that we are at the beginning of a sequence of outdegrees). This parameter will not be modified. 
	 * @param outd if <code>window</code> is not <code>null</code>, this is an array with as many elements
	 *   as {@link #windowSize}; <code>outd[(x-i) mod offsetStep]</code> contains the outdegree of node <code>x</code>
	 *   &minus;<code>i</code> for <code>i</code> greater than 0; at the end, this will be true also for <code>i</code> equal to 0.
	 * @param blockOutdegrees if <code>window</code> is not <code>null</code>, an array of size {@link #offsetStep} 
	 *   containing the outdegrees of the nodes the block to which <code>x</code> belongs, if <code>x</code> is
	 *   not a the start of a block; otherwise, the content is immaterial. After the call, in the first case the vector
	 *   will be left unchanged, whereas in the second case it will be filled with the outdegrees of the nodes of the block
	 *   to which <code>x</code> belongs.
	 * @return an iterator over the successors of <code>x</code>.
	 * @throws IllegalStateException if <code>window</code> is <code>null</code> and {@link #offsetStep} is 0.
	 * 	 
	 */
	protected IntIterator successors( final int x, final InputBitStream ibs, final int window[][], final int outd[], final int blockOutdegrees[] ) throws IllegalStateException {
		int i;
		int extraCount;
		final int ref, refIndex;
		int block[] = null, blockCount = 0;
		int left[] = null;
		int len[] = null;

		if (x < 0 || x >= n) throw new IllegalArgumentException( "Node index out of range:" + x );

		try {
			final int d;
			final int cyclicBufferSize = windowSize + 1;
			
			if ( window == null ) { // We position on the given node.
				// Without offsets, we just give up.
				if ( offsetStep <= 0 ) throw new IllegalStateException( "You cannot position a stream without offsets" );
				// If window is null, we use the position method, which may modify the offset/outdegree caches.
				d = position( ibs, x );
			}
			else {
				if ( offsetStep <= 1 ) d = outd[ x % cyclicBufferSize ] = readOutdegree( ibs ); // We just read the outdegree.
				else {
					// We are reading a rearranged graph file.
					if ( x % offsetStep == 0 ) {
						// We are at the start of a block, so we read and cache the outdegrees.
						final int actualStep = Math.min( offsetStep, n - x );
						for( i = 0; i < actualStep; i++ ) blockOutdegrees[ i ] = readOutdegree( ibs );
					}
					// We fetch the current outdegree from the block cache.
					d = outd[ x % cyclicBufferSize ] = blockOutdegrees[ x % offsetStep ];
				}
			}
			
			if ( d == 0 ) return Iterators.EMPTY_ITERATOR;
			
			// We read the reference only if the actual window size is larger than one (i.e., the one specified by the user is larger than 0).
			if ( windowSize > 0 ) ref = readReference( ibs );
			else ref = -1;
			refIndex = ( x - ref + cyclicBufferSize ) % cyclicBufferSize; // The index in window[] of the node we are referring to (it makes sense only if ref>0).

			if ( ref > 0 ) { // This catches both no references at all and no reference specifically for this node.
				if ( ( blockCount = readBlockCount( ibs ) ) !=  0 ) block = new int[ blockCount ];

				int copied = 0, total = 0; // The number of successors copied, and the total number of successors specified in some copy block.
				for( i = 0; i < blockCount; i++ ) {
					block[ i ] = readBlock( ibs ) + ( i == 0 ? 0 : 1 );
					total += block[ i ];
					if ( i % 2 == 0 ) copied += block[ i ];
				}
				// If the block count is even, we must compute the number of successors copied implicitly.
				if ( blockCount % 2 == 0 ) copied += ( window != null ? outd[ refIndex ] : outdegree( x - ref ) ) - total;
				extraCount = d - copied;
			}
			else extraCount = d;
			
			int intervalCount = 0; // Number of intervals

			if ( extraCount > 0 ) {

				// Prepare to read intervals, if any
				if ( minIntervalLength != NO_INTERVALS && ( intervalCount = ibs.readGamma() ) != 0 ) {

					int prev = 0; // Holds the last integer in the last interval.
					left = new int[ intervalCount ];
					len = new int[ intervalCount ];

					// Now we read intervals
					left[ 0 ] = prev = Fast.nat2int( ibs.readGamma() ) + x;
					len[ 0 ] = ibs.readGamma() + minIntervalLength;

					prev += len[ 0 ];
					extraCount -= len[ 0 ];

					for ( i = 1; i < intervalCount; i++ ) {
						left[ i ] = prev = ibs.readGamma() + prev + 1;
						len[ i ] = ibs.readGamma() + minIntervalLength;
						prev += len[ i ];
						extraCount -= len[ i ];
					}
				}
			}
				
			final int residualCount = extraCount; // Just to be able to use an anonymous class.

			final IntIterator residualIterator = residualCount == 0 ? null : new AbstractIntIterator() {
					int prev = -1, i = residualCount;
					
					public boolean hasNext() { return i != 0; };
					
					public int nextInt() { 
						try { 
							if ( ! hasNext() ) throw new java.util.NoSuchElementException();
							i--;
							if ( prev == -1 ) return prev = x + Fast.nat2int( readResidual( ibs ) );
							else return prev = readResidual( ibs ) + prev + 1;
						} 
						catch( IOException cantHappen ) { throw new RuntimeException( cantHappen ); } 
					}
				};
			
			// The extra part is made by the contribution of intervals, if any, and by the residuals iterator.
			final IntIterator extraIterator = intervalCount == 0 
				? residualIterator 
				: ( residualCount == 0 
					? (IntIterator)new IntIntervalSequenceIterator( left, len )
					: (IntIterator)new MergedIntIterator( new IntIntervalSequenceIterator( left, len ), residualIterator )
					);

			final IntIterator blockIterator = ref <= 0
				? null 
				: new MaskedIntIterator(
										// ...block for masking copy and...
										block, 
										// ...the reference list (either computed recursively or stored in window)...
										window != null 
										? IntIterators.wrap( window[ refIndex ], 0, outd[ refIndex ] ) 
										: successors( x - ref ) // This is the recursive lazy part of the construction.
										);
			
			if ( ref <= 0 ) return extraIterator;
			else return extraIterator == null
					 ? blockIterator
					 : (IntIterator)new MergedIntIterator( blockIterator, extraIterator, d );
			
		}
		catch( IOException cantHappen ) { System.err.println("Accessing node " + x ); throw new RuntimeException( cantHappen ); }
		
	}

	
	private class BVGraphNodeIterator extends NodeIterator {
		final private int n = numNodes();
		/** Our bit stream. */
		final InputBitStream ibs;
		/** We keep the size of the cyclic buffer (the window size + 1 ) in a local variable. */
		final private int cyclicBufferSize = windowSize + 1;
		/** At any time, window will be ready to be passed to {@link BVGraph#successors(int, InputBitStream, int[][], int[], int[])} */ 
		final private int window[][] = new int[ cyclicBufferSize ][ INITIAL_SUCCESSOR_LIST_LENGTH ];
		/** At any time, outd will be ready to be passed to {@link BVGraph#successors(int, InputBitStream, int[][], int[], int[])} */ 
		final private int outd[] = new int[ cyclicBufferSize ];
		/** At any time, blockOutdegrees will be ready to be passed to {@link BVGraph#successors(int, InputBitStream, int[][], int[], int[])} */ 
		final private int blockOutdegrees[] = offsetStep > 1 ? new int[ offsetStep ] : null;
		/** The index of the node from which we started iterating. */
		final private int from;
		/** The index of the node just before the next one. */
		private int curr;

		public BVGraphNodeIterator( final InputBitStream ibs, final int from ) {
			this.from = from;
			this.ibs = ibs;
			if ( from != 0 ) {
				if ( offsetStep <= 0 ) throw new IllegalStateException( "You cannot iterate from a chosen node without offsets" );
				position( ibs, from );
			}
			curr = from - 1;
		}

		/** At each call, we build the successor iterator (making a call to {@link BVGraph#successors(int, InputBitStream, int[][], int[], int[])},
		 *  and we completely iterate over it, filling the appropriate entry in <code>window</code>.
		 */
		public int nextInt() {
			if ( ! hasNext() ) throw new NoSuchElementException();

			final int currIndex = ++curr % cyclicBufferSize;
			final IntIterator i = BVGraph.this.successors( curr, ibs, window, outd, blockOutdegrees );

			if ( window[ currIndex ].length < outd[ currIndex ] ) window[ currIndex ] = new int[ outd[ currIndex ] ];
			IntIterators.unwrap( i, window[ currIndex ] );

			return curr;
		}

		public boolean hasNext() {
			return ( curr < n - 1 );
		}

		public IntIterator successors() {
			if ( curr == from - 1 ) throw new IllegalStateException();

			final int currIndex = curr % cyclicBufferSize;
			return IntIterators.wrap( window[ currIndex ], 0, outd[ currIndex ] );
		}

		public int[] successorArray() {
			if ( curr == from - 1 ) throw new IllegalStateException();	

			return window[ curr % cyclicBufferSize ];
		}

		public int outdegree() {
			if ( curr == from - 1 ) throw new IllegalStateException();
			return outd[ curr % cyclicBufferSize ];
		}
	};


	/** This method returns a node iterator for scanning the graph sequentially, starting from the given node.
	 *  It keeps track of a sliding window of {@link #windowSize()} previous successor lists
	 *  to speed up the iteration of graphs with significant referentiation.
	 * 
	 *  @param from the node from which the iterator will iterate.
	 *  @return a {@link NodeIterator} for accessing nodes and successors sequentially.
	 */

	public NodeIterator nodeIterator( final int from ) {
		try {
			return offsetStep == -1 
				? new BVGraphNodeIterator( new InputBitStream( new FileInputStream( basename + ".graph" ), STD_BUFFER_SIZE ), from )
				: new BVGraphNodeIterator( isMemory ? new InputBitStream( graphMemory ) : new InputBitStream( new FastMultiByteArrayInputStream( graphStream ) ), from );
		} catch ( FileNotFoundException e ) {
			throw new IllegalStateException( "The graph file \"" + basename + ".graph\" cannot be found" );
		}
	}
	
	/** A reasonable format for real numbers. */
	private static java.text.NumberFormat formatDouble = new java.text.DecimalFormat( "#.00" );
	
	/** Formats a number.
	 *
	 * <P>This method formats a double printing just two fractional digits.
	 *
	 * @param d a number.
	 * @return a string containing a pretty print of the number.
	 */
	private static String format( final double d ) {
		final StringBuffer s = new StringBuffer();
		return formatDouble.format( d, s, new java.text.FieldPosition( 0 ) ).toString();
	}

	/* The following private methods handle the flag mask. They are the only methods which replicate
	 * the shifting logic specified in the flag-mask definition.
	 */

	/** Sets the {@link #flags} attribute to the given value, and updates appropriately the 
	 *  individual coding attributes (<code>g&hellip;Coding</code>).
	 *  
	 *  <P>If a certain bit-slot within <code>flags</code> is not specified (i.e., 0) the corresponding 
	 *  coding variable is left unchanged, making the assumption that it is the default value (this condition
	 *  is anyway not checked for).
	 * 	  
	 * @param flags a mask of flags as specified by the constants of this class.
	 */
	private void setFlags( int flags ) {
		this.flags = flags;
		if ( ( flags & 0xF ) != 0 ) outdegreeCoding = flags & 0xF;
		if ( ( ( flags >>> 4 ) & 0xF ) != 0 ) blockCoding = ( flags >>> 4 ) & 0xF;
		if ( ( ( flags >>> 8 ) & 0xF ) != 0 ) residualCoding = ( flags >>> 8 ) & 0xF;
		if ( ( ( flags >>> 12 ) & 0xF ) != 0 ) referenceCoding = ( flags >>> 12 ) & 0xF;
		if ( ( ( flags >>> 16 ) & 0xF ) != 0 ) blockCountCoding = ( flags >>> 16 ) & 0xF;
		if ( ( ( flags >>> 20 ) & 0xF ) != 0 ) offsetCoding = ( flags >>> 20 ) & 0xF;
	}

	/** Produces a string representing the values coded in the given flag mask.
	 * 
	 * @param flags a flag mask.
	 * @return a string representing the flag mask.
	 */
	private static MutableString flags2String( int flags ) {
		MutableString s = new MutableString();

		if ( ( flags & 0xF ) != 0 ) s.append( " | " ).append( "OUTDEGREES_" ).append( CompressionFlags.CODING_NAME[ flags & 0xF ] );
		if ( ( ( flags >>> 4 ) & 0xF ) != 0 ) s.append( " | " ).append( "BLOCKS_" ).append( CompressionFlags.CODING_NAME[ ( flags >>> 4 ) & 0xF ] );
		if ( ( ( flags >>> 8 ) & 0xF ) != 0 ) s.append( " | " ).append( "RESIDUALS_" ).append( CompressionFlags.CODING_NAME[ ( flags >>> 8 ) & 0xF ] );
		if ( ( ( flags >>> 12 ) & 0xF ) != 0 ) s.append( " | " ).append( "REFERENCES_" ).append( CompressionFlags.CODING_NAME[ ( flags >>> 12 ) & 0xF ] );
		if ( ( ( flags >>> 16 ) & 0xF ) != 0 ) s.append( " | " ).append( "BLOCK_COUNT_" ).append( CompressionFlags.CODING_NAME[ ( flags >>> 16 ) & 0xF ] );
		if ( ( ( flags >>> 20 ) & 0xF ) != 0 ) s.append( " | " ).append( "OFFSETS_" ).append( CompressionFlags.CODING_NAME[ ( flags >>> 20 ) & 0xF ] );

		if ( s.length() != 0 ) s.delete( 0, 3 );
		return s;
	}

	/** Produces a flag mask corresponding to a given string.
	 * 
	 * @param flagString a flag string.
	 * @return the flag mask.
	 * @throws IOException if the flag string is malformed.
	 */
	private static int string2Flags( String flagString ) throws IOException {
		int flags = 0;
		
		if ( flagString != null && flagString.length() != 0 ) {
			String f[] = flagString.split( "\\|" );
			for( int i = 0; i < f.length; i++ ) {
				try {
					flags |= BVGraph.class.getField( f[ i ].trim() ).getInt( BVGraph.class );
				}
				catch ( Exception notFound ) {
					throw new IOException( "Compression flag " + f[ i ] + " unknown." );
				}
			}
		}
		return flags;
	}


	/** Creates a new {@link BVGraph} by loading a compressed graph file from disk to memory.
	 *
	 * @param basename the basename of the graph.
	 * @param offsetStep the desired offset step (0 means that we do not want to load offsets at all, -1 that
	 * the we do not want even load the graph file).
	 * @param pm a progress meter used while loading the graph, or <code>null</code>.
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */
	public static ImmutableGraph load( CharSequence basename, int offsetStep, ProgressMeter pm ) throws IOException {
		return new BVGraph().loadInternal( basename, offsetStep, pm );
	}

	/** Creates a new {@link BVGraph} by loading a compressed graph file from disk to memory, with no progress meter.
	 *
	 * @param basename the basename of the graph.
	 * @param offsetStep the desired offset step (0 means that we do not want to load offsets at all, -1 that
	 * the we do not want even load the graph file).
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */
	public static ImmutableGraph load( CharSequence basename, int offsetStep ) throws IOException {
		return BVGraph.load( basename, offsetStep, null );
	}

	/** Creates a new {@link BVGraph} by loading a compressed graph file from disk to memory, with no progress meter and
	 *  all offsets.
	 *
	 * @param basename the basename of the graph.
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */
	public static ImmutableGraph load( CharSequence basename ) throws IOException {
		return BVGraph.load( basename, 1 );
	}

	/** Creates a new {@link BVGraph} by loading a compressed graph file from disk to memory, with 
	 *  all offsets.
	 *
	 * @param basename the basename of the graph.
	 * @param pm a progress meter used while loading the graph, or <code>null</code>.
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */
	public static ImmutableGraph load( CharSequence basename, ProgressMeter pm ) throws IOException {
		return BVGraph.load( basename, 1, pm );
	}

	/** Creates a new {@link BVGraph} by loading a compressed graph file from disk to memory, with no progress meter and
	 * without offsets.
	 * 
	 * @param basename the basename of the graph.
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */
	public static ImmutableGraph loadSequential( CharSequence basename ) throws IOException {
		return BVGraph.loadSequential( basename, null );
	}

	/** Creates a new {@link BVGraph} by loading a compressed graph file from disk to memory, without offsets. 
	 *
	 * @param basename the basename of the graph.
	 * @param pm a progress meter used while loading the graph, or <code>null</code>.
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */
	public static ImmutableGraph loadSequential( CharSequence basename, ProgressMeter pm ) throws IOException {
		return BVGraph.load( basename, 0, pm );
	}



	/** Creates a new {@link BVGraph} by loading just the metadata of a compressed graph file.
	 * 
	 * @param basename the basename of the graph.
	 * @param pm a progress meter.
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the metadata.
	 */
	public static ImmutableGraph loadOffline( CharSequence basename, ProgressMeter pm ) throws IOException {
		return BVGraph.load( basename, -1, pm );
	}

	/** Creates a new {@link BVGraph} by loading just the metadata of a compressed graph file.
	 * 
	 * @param basename the basename of the graph.
	 * @return a {@link BVGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the metadata.
	 */
	public static ImmutableGraph loadOffline( CharSequence basename ) throws IOException {
		return BVGraph.loadOffline( basename, null );
	}

	/** Loads a compressed graph file from disk into this graph. Note that this method should
	 *  be called <em>only</em> on a newly created graph.
	 *
	 * @param basename the basename of the graph.
	 * @param offsetStep the desired offset step (0 means that we do not want to load offsets at all).
	 * @param pm a progress meter used while loading the graph, or <code>null</code>.
	 * @return this graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */
	protected BVGraph loadInternal( CharSequence basename, int offsetStep, ProgressMeter pm ) throws IOException {
		int i;

		this.offsetStep = offsetStep;
		this.basename = new MutableString( basename );

		// First of all, we read the property file to get the relevant data.
		final FileInputStream propertyFile = new FileInputStream( basename + ".properties" );
		final Properties properties = new Properties();
		properties.load( propertyFile );
		propertyFile.close();

		// We parse the properties and perform some consistency check and assignments.
		setFlags( string2Flags( properties.getProperty( "compressionflags" ) ) );
		if ( properties.getProperty( "version" ) == null ) throw new IOException( "Missing format version information" );
		else if ( Integer.parseInt( properties.getProperty( "version" ) ) > BVGRAPH_VERSION ) throw new IOException( "This graph uses format " + properties.getProperty( "version" ) + ", but this class can understand only graphs up to format " + BVGRAPH_VERSION );;
		n = Integer.parseInt( properties.getProperty( "nodes" ) );
		m = Long.parseLong( properties.getProperty( "arcs" ) );
		windowSize = Integer.parseInt( properties.getProperty( "windowsize" ) );
		maxRefCount = Integer.parseInt( properties.getProperty( "maxrefcount" ) );
		minIntervalLength = Integer.parseInt( properties.getProperty( "minintervallength" ) );
		if ( properties.getProperty( "zetak" ) != null ) zetaK = Integer.parseInt( properties.getProperty( "zetak" ) );
		// Soft check due to previous usage of toString() instead of getName()
		if ( ! properties.getProperty( "graphclass" ).endsWith( this.getClass().getName() ) ) throw new IOException( "This class (" + this.getClass().getName() + ") cannot load a graph stored using " + properties.getProperty( "graphclass" ) );

			
		InputBitStream offsetIbs = null;

		if ( offsetStep > 0 ) offsetIbs =  new InputBitStream( new FileInputStream( basename + ".offsets" ), STD_BUFFER_SIZE );

		if ( offsetStep == 0 || offsetStep == 1 ) {
			// No permutation is required: the graph file is loaded as such
			final FileInputStream fis = new FileInputStream( basename + ".graph" );

			// read the whole graph into memory
			if ( pm != null ) {
				System.err.print( "Loading graph..." );
				pm.itemsName( "bytes" );
				pm.start();
			}

			if ( fis.getChannel().size() <= Integer.MAX_VALUE ) {
				graphMemory = new byte[ ( int ) fis.getChannel().size() ];
				fis.read( graphMemory );
				fis.close();
				isMemory = true;
			}
			else graphStream = new FastMultiByteArrayInputStream( fis, fis.getChannel().size() );

			if ( pm != null ) {
				pm.stop();
				pm.count( isMemory ? graphMemory.length : graphStream.length );
				System.err.println( " done. " + pm );
			}

			if ( offsetStep == 1 ) {
				// read offsets, if required

				offset = new long[ n + 1 ];

				if ( pm != null ) {
					System.err.print( "Loading offsets..." );
					pm.itemsName( "deltas" );
					pm.start();
				}
				
				long off = 0;
				for( i = 0; i <= n; i++ ) offset[ i ] = off = readOffset( offsetIbs ) + off;
				
				if ( pm != null ) {
					pm.stop();
					pm.count( offset.length );
					System.err.println( " done. " + pm );
				}
			}
		}
		else if ( offsetStep > 1 ) {
			isMemory = true;

			// We must permute the graph file, so to load offsets only partially 	
			final File graphFile = new File( basename + ".graph" );
			if ( graphFile.length() > Integer.MAX_VALUE ) throw new IOException( "You cannot use a positive offset with a file larger than 2 GiB" );
			final InputBitStream graphIbs = new InputBitStream( new FileInputStream( graphFile ), STD_BUFFER_SIZE );

			offset = new long[ ( n + offsetStep - 1 ) / offsetStep + 1 ];
			final FastByteArrayOutputStream cacheFbaos = new FastByteArrayOutputStream();
			final OutputBitStream cache = new OutputBitStream( cacheFbaos, 0 );
			final OutputBitStream graphObs = new OutputBitStream( graphMemory = new byte[ (int)graphFile.length() ] );
			/* A buffer used to transfer the content of a given successor list (excluding the outdegree) to the cache.
			 * This will be superseeded sooner or later by some direct-transfer method.
			 */
			final byte buffer[] = new byte[ 1024 * 1024 ];
			
			// read the whole graph into memory rechunking it in blocks of offsetStep nodes.
			if ( pm != null ) {
				System.err.print( "Loading graph..." );
				pm.itemsName( "nodes" );
				pm.start();
			}

			int j = 0, delta, outLen, cacheLen;

			delta = readOffset( offsetIbs );
			if ( delta != 0 )
				throw new IOException( "The first offset should always be zero, but it was " + delta + " instead" );
				
			long currOffset = delta;
			offset[ j++ ] = currOffset;

			for( i = 1; i <= n; i++ ) {
				// We read the offset delta relative to node i
				delta = readOffset( offsetIbs );
				currOffset += delta;
				// We copy the outdegree of node i-1 directly to the graph bitstream
				outLen = writeOutdegree( graphObs, readOutdegree( graphIbs ) );
				// We copy the remaining part of the entry relative to node i-1 to the cache
				graphIbs.read( buffer, delta - outLen );
				cache.write( buffer, delta - outLen );

				if ( i % offsetStep == 0 ) {
					// If needed, we flush the cache onto the graph bitstream
					cacheLen = (int)cache.writtenBits();
					cache.flush();
					graphObs.write( cacheFbaos.array, cacheLen ); // when i = 0, cacheLen = 0.
					cacheFbaos.reset();
					cache.writtenBits( 0 );
					offset[ j++ ] = currOffset;
				}

				if ( pm != null ) pm.update();
			}

			if ( n % offsetStep != 0 ) {
				// We need one more flush if n is not a multiple of offsetStep
				cacheLen = (int)cache.writtenBits();
				cache.flush();
				graphObs.write( cacheFbaos.array, cacheLen ); // when i = 0, cacheLen = 0.
				offset[ j++ ] = currOffset;
			}

			graphObs.flush();

			if ( pm != null ) {
				pm.stop();
				System.err.println( " done. " + pm );
			}

			cache.close();
			graphIbs.close();
		}

		if ( offsetIbs != null ) offsetIbs.close();

		// We finally create the outdegreeIbs and, if needed, the two caches
		if ( offsetStep >= 0 ) outdegreeIbs = isMemory ? new InputBitStream( graphMemory ): new InputBitStream( new FastMultiByteArrayInputStream( graphStream ) );

		if ( offsetStep > 1 ) {
			outdCache = new int[ offsetStep ];
			offsetCache = new long[ offsetStep ];
		}

		return this;
	}



	/** This method tries to express an increasing sequence of natural numbers <code>x</code> as a union of an increasing
	 *  sequence of intervals and an increasing sequence of residual elements. More precisely, this intervalization works
	 *  as follows: first, one looks at <code>x</code> as a sequence of intervals (i.e., maximal sequences of consecutive
	 *  elements); those intervals whose length is &ge; <code>minInterval</code> are stored in the lists <code>left</code>
	 *  (the list of left extremes) and <code>len</code> (the list of lengths; the length of an integer interval is the
	 *  number of integers in that interval). The remaining integers, called <em>residuals</em> are stored in the
	 *  <code>residual</code> list.
	 * 
	 *  <P>Note that the previous content of <code>left</code>, <code>len</code> and <code>residual</code> is lost.
	 *
	 *  @param x the list to be intervalized (an increasing list of natural numbers).
	 *  @param minInterval the least length that a maximal sequence of consecutive elements must have in order for it to
	 *   be considered as an interval.
	 *  @param left the resulting list of left extremes of the intervals.
	 *  @param len the resulting list of interval lengths.
	 *  @param residuals the resulting list of residuals.
	 *  @return the number of intervals.
	 */
	protected static int intervalize( IntArrayList x, int minInterval, IntArrayList left, IntArrayList len, IntArrayList residuals ) {
		int nInterval = 0;
		int vl = x.size();
		int v[] = x.elements();
		int i, j;

		left.clear(); len.clear(); residuals.clear();
		for( i = 0; i < vl; i++ ) {
			j = 0;
			if ( i < vl - 1 && v[ i ] + 1 == v[ i + 1 ] ) {
				do j++; while( i + j < vl - 1 && v[ i + j ] + 1 == v[ i + j + 1 ] );
				j++;
				// Now j is the number of integers in the interval.
				if ( j >= minInterval ) {
					left.add( v[ i ] );
					len.add( j );
					nInterval++;
					i += j - 1;
				}
			}
			if ( j < minInterval ) residuals.add( v[ i ] );
		}
		return nInterval;
	}




	/** Scratch variables used by the {@link #diffComp(OutputBitStream, int, int, int[], int, int[], int, boolean)} method. */
	private IntArrayList extras = new IntArrayList(), blocks = new IntArrayList(), residuals = new IntArrayList(),
		left = new IntArrayList(), len = new IntArrayList();

	/** Compresses differentially the given list. This method is given a node (with index <code>currNode</code>) called the
	 * current node, with its successor list (contained in the array <code>currList[0..currLen-1]</code>), and another node
	 * (with index <code>currNode</code>&minus;<code>ref</code>), called the reference node, with its successor list (contained in the array
	 * <code>refList[0..refLen-1]</code>). This method produces, onto the given output bit stream, the compressed successor
	 * list of the current node using the reference node given (except for the outdegree); the number of bits written is returned. 
	 *
	 * Note that <code>ref</code> may be zero, in which case no differential compression is made.
	 *
	 * @param obs an output bit stream where the compressed data will be stored.
	 * @param currNode the index of the node this list of outlinks refers to.
	 * @param ref the distance from the reference list.
	 * @param refList the reference list.
	 * @param refLen the length of the reference list.
	 * @param currList the current list.
	 * @param currLen the current list length.
	 * @param forReal if true, we are really writing data (i.e., <code>obs</code> is not just a bit count stream).
	 * @return the number of bits written.
	 */
	private int diffComp( OutputBitStream obs, int currNode, int ref, int refList[], int refLen, int currList[], int currLen, boolean forReal ) throws IOException {
		// Bits already written onto the output bit stream
		final long writtenBitsAtStart = obs.writtenBits();

		// We build the list of blocks copied and skipped (alternatively) from the previous list.
		int i, j = 0, k = 0, prev = 0, currBlockLen = 0;
		boolean copying = true;

		if ( ref == 0 ) refLen = 0; // This guarantees that we will not try to differentially compress when ref == 0.

		extras.clear();
		blocks.clear();

		// j is the index of the next successor of the current node we must examine
		// k is the index of the next successor of the reference node we must examine
		// copying is true iff we are producing a copy block (instead of an ignore block)
		// currBlockLen is the number of entries (in the reference list) we have already copied/ignored (in the current block)
		while( j < currLen && k < refLen ) {
			if ( copying ) { // First case: we are currectly copying entries from the reference list
				if ( currList[ j ] > refList[ k ] ) {
					/* If while copying we trespass the current element of the reference list,
					   we must stop copying. */
					blocks.add( currBlockLen );
					copying = false;
					currBlockLen = 0;
				}
				else if ( currList[ j ] < refList[ k ] ) {
					/* If while copying we find a non-matching element of the reference list which
					   is larger than us, we can just add the current element to the extra list
					   and move on. j gets increased. */
					extras.add( currList[ j++ ] );
				}
				else { // currList[ j ] == refList[ k ]
					/* If the current elements of the two lists are equal, we just increase the block length. 
					   both j and k get increased. */
					j++;
					k++;
					currBlockLen++;
				}
			}
			else { // Second case: we are currently skipping entries from the reference list
				if ( currList[ j ] < refList[ k ] ) {
					/* If we did not trespass the current element of the reference list, we just
					   add the current element to the extra list and move on. j gets increased. */
					extras.add( currList[ j++ ] );
				}
				else if ( currList[ j ] > refList[ k ] ) {
					/* If we trespassed the currented element of the reference list, we
					   increase the block length. k gets increased. */
					k++;
					currBlockLen++;
				}
				else { // currList[ j ] == refList[ k ]
					/* If we found a match we flush the current block and start a new copying phase. */
					blocks.add( currBlockLen );
					copying = true;
					currBlockLen = 0;
				}
			}
		}
		
		/* We do not record the last block. The only case when we have to enqueue the last block's length
		 * is when we were copying and we did not copy up to the end of the reference list.
		 */
		if ( copying && k < refLen ) blocks.add( currBlockLen );
	
		// If there are still missing elements, we add them to the extra list.
		while( j < currLen ) extras.add( currList[ j++ ] );

		// We store locally the resulting arrays for faster access.
		final int block[] = blocks.elements(), blockCount = blocks.size(), extraCount = extras.size();
	
		// If we have a nontrivial reference window we write the reference to the reference list.
		if ( windowSize > 0 ) 
			writeReference( obs, ref );

		if ( STATS && forReal ) referenceStats.println( ref );

		// Then, if the reference is not void we write the length of the copy list.
		if ( ref != 0 ) {
			writeBlockCount( obs, blockCount );
			
			if ( STATS && forReal ) blockCountStats.println( blockCount );

			// Then, we write the copy list; all lengths except the first one are decremented.
			if ( blockCount > 0 ) {
				writeBlock( obs, block[ 0 ] );
				for( i = 1; i < blockCount; i++ ) 
					writeBlock( obs, block[ i ] - 1 );

				if ( STATS && forReal ) {
					blockStats.println( block[ 0 ] ); 
					for( i = 1; i < blockCount; i++ ) blockStats.println( block[ i ] - 1 );
				}
			}
		}

		// Finally, we write the extra list.
		if ( extraCount > 0 ) {

			final int residual[], residualCount;

			if ( minIntervalLength != NO_INTERVALS ) {
				// If we are to produce interval, we first compute them.
				final int intervalCount = intervalize( extras, minIntervalLength, left, len, residuals );
				
				// We write the number of intervals.
				obs.writeGamma( intervalCount );
				
				if ( STATS && forReal ) intervalCountStats.println( intervalCount );
				
				// We write out the intervals.
				for( i = 0; i < intervalCount; i++ ) {
					if ( i == 0 ) obs.writeGamma( Fast.int2nat( ( prev = left.getInt( i ) ) - currNode ) );
					else obs.writeGamma( left.getInt( i ) - prev - 1 );
					prev = left.getInt( i ) + len.getInt( i );
					obs.writeGamma( len.getInt( i ) - minIntervalLength );
				}
				
				if ( STATS && forReal ) 
					for( i = 0; i < intervalCount; i++ ) {
					if ( i == 0 ) leftStats.println( Fast.int2nat( ( prev = left.getInt( i ) ) - currNode ) );
					else leftStats.println( left.getInt( i ) - prev - 1 );
					prev = left.getInt( i ) + len.getInt( i );
					lenStats.println( len.getInt( i ) - minIntervalLength );
				}
				
				
				residual = residuals.elements();
				residualCount = residuals.size();
			}
			else {
				residual = extras.elements();
				residualCount = extras.size();
			}
					 
			if ( STATS && forReal ) residualCountStats.println( residualCount );

			// Now we write out the residuals, if any
			if ( residualCount != 0 ) {
				writeResidual( obs, Fast.int2nat( ( prev = residual[ 0 ] ) - currNode ) );
				for( i = 1; i < residualCount; i++ ) {
					if ( residual[ i ] == prev ) throw new IllegalArgumentException( "Repeated successor " + prev + " in successor list of node " + currNode );
					writeResidual( obs, residual[ i ] - prev - 1 );
					prev = residual[ i ];
				}
			
				if ( STATS && forReal ) {
					residualStats.println( Fast.int2nat( ( prev = residual[ 0 ] ) - currNode ) );
					for( i = 1; i < residualCount; i++ ) {
						residualStats.println( residual[ i ] - prev - 1 );
						prev = residual[ i ];
					}
				}
			}
			
		}

		return (int)( obs.writtenBits() - writtenBitsAtStart );
	}

	/** Writes the given graph using a given base name.
	 *
	 * @param graph a graph to be compressed.
	 * @param basename a base name.
	 * @param windowSize the window size (-1 for the default value).
	 * @param maxRefCount the maximum reference count (-1 for the default value).
	 * @param minIntervalLength the minimum interval length (-1 for the default value).
	 * @param zetaK the parameter used for residual &zeta;-coding, if used (-1 for the default value).
	 * @param flags the flag mask.
	 * @param pm a progress meter to measure the state of compression, or <code>null</code> if no metering is required.
	 * @throws IOException if some exception is raised while writing the graph.
	 */
	public static void store( ImmutableGraph graph, CharSequence basename, int windowSize, int maxRefCount, int minIntervalLength, 
		int zetaK, int flags, ProgressMeter pm ) throws IOException {
		BVGraph g = new BVGraph();
		if ( windowSize != -1 ) g.windowSize = windowSize;
		if ( maxRefCount != -1 ) g.maxRefCount = maxRefCount;
		if ( minIntervalLength != -1 ) g.minIntervalLength = minIntervalLength;
		if ( zetaK != -1 ) g.zetaK = zetaK;
		g.setFlags( flags );
		g.storeInternal( graph, basename, pm );
	}

	/** Writes the given graph using a given base name, without any progress meter.
	 *
	 * @param graph a graph to be compressed.
	 * @param basename a base name.
	 * @param windowSize the window size (-1 for the default value).
	 * @param maxRefCount the maximum reference count (-1 for the default value).
	 * @param minIntervalLength the minimum interval length (-1 for the default value).
	 * @param zetaK the parameter used for residual &zeta;-coding, if used (-1 for the default value).
	 * @param flags the flag mask.
	 * @throws IOException if some exception is raised while writing the graph.
	 */
	public static void store( ImmutableGraph graph, CharSequence basename, int windowSize, int maxRefCount, int minIntervalLength, 
		int zetaK, int flags ) throws IOException {
		BVGraph.store( graph, basename, windowSize, maxRefCount, minIntervalLength, zetaK, flags, null );
	}

	/** Writes the given graph using a given base name, with all
	 * parameters set to their default values.
	 *
	 * @param graph a graph to be compressed.
	 * @param basename a base name.
	 * @param pm a progress meter to measure the state of compression, or <code>null</code> if no metering is required.
	 * @throws IOException if some exception is raised while writing the graph.
	 */
	public static void store( ImmutableGraph graph, CharSequence basename, ProgressMeter pm ) throws IOException {
		BVGraph.store( graph, basename, -1, -1, -1, -1, 0, pm );
	}

	/** Writes the given graph using a given base name, without any progress meter and with all
	 * parameters set to their default values.
	 *
	 * @param graph a graph to be compressed.
	 * @param basename a base name.
	 * @throws IOException if some exception is raised while writing the graph.
	 */
	public static void store( ImmutableGraph graph, CharSequence basename ) throws IOException {
		BVGraph.store( graph, basename, null );
	}


	/** Writes the given graph <code>graph</code> using a given base name, and the compression parameters and flags
	 * of this graph object. Note that the latter is relevant only as far as parameters and flags are concerned; its
	 * content is really irrelevant.
	 *
	 * @param graph a graph to be compressed.
	 * @param basename a base name.
	 * @param pm a progress meter to measure the state of compression, or <code>null</code> if no metering is required.
	 * @throws IOException if some exception is raised while writing the graph.
	 */
	private void storeInternal( ImmutableGraph graph, CharSequence basename, ProgressMeter pm ) throws IOException {
		// Used for differential compression
		final OutputBitStream bitCount = new OutputBitStream( NullOutputStream.getInstance(), 0  );
		int outd, currNode, currIndex, j, best, bestIndex, cand, t = 0, n = graph.numNodes();
		long bitOffset = 0;

		OutputBitStream graphObs = new OutputBitStream( new FileOutputStream( basename + ".graph" ), STD_BUFFER_SIZE );
		OutputBitStream offsetObs = new OutputBitStream( new FileOutputStream( basename + ".offsets" ), STD_BUFFER_SIZE );

		if ( STATS ) {
			offsetStats = new PrintWriter( new FileWriter( basename + ".offsetStats" ) );
			referenceStats = new PrintWriter( new FileWriter( basename + ".referenceStats" ) );
			outdegreeStats = new PrintWriter( new FileWriter( basename + ".outdegreeStats" ) );
			blockCountStats = new PrintWriter( new FileWriter( basename + ".blockCountStats" ) );
			blockStats = new PrintWriter( new FileWriter( basename + ".blockStats" ) );
			intervalCountStats = new PrintWriter( new FileWriter( basename + ".intervalCountStats" ) );
			leftStats = new PrintWriter( new FileWriter( basename + ".leftStats" ) );
			lenStats = new PrintWriter( new FileWriter( basename + ".lenStats" ) );
			residualCountStats = new PrintWriter( new FileWriter( basename + ".residualCountStats" ) );
			residualStats = new PrintWriter( new FileWriter( basename + ".residualStats" ) );
		}

		final int cyclicBufferSize = windowSize + 1;
		// Cyclic array of previous lists.
		int list[][] = new int[ cyclicBufferSize ][ INITIAL_SUCCESSOR_LIST_LENGTH ];
		// For each list, its length.
		int listLen[] = new int[ cyclicBufferSize ];
		// For each list, the depth of its references.
		int refCount[] = new int[ cyclicBufferSize ];

		long totRef = 0, totDist = 0, totLinks = 0;

		if ( pm != null ) {
			System.err.print( "Storing..." );
			pm.itemsName( "nodes" );
			pm.expectedUpdates( n );
			pm.start();
		}

		// We iterate over the nodes of graph
		for ( NodeIterator nodeIterator = graph.nodeIterator(); nodeIterator.hasNext(); ) {
			// currNode is the currently examined node, of outdegree outd, with index currIndex (within the cyclic array)
			currNode = nodeIterator.nextInt();
			outd = nodeIterator.outdegree();// get the number of successors of currNode
			currIndex = currNode % cyclicBufferSize;

			// We write the current offset to the offset stream
			writeOffset( offsetObs, (int)( graphObs.writtenBits() - bitOffset ) );

			if ( STATS ) offsetStats.println( graphObs.writtenBits() - bitOffset );

			bitOffset = graphObs.writtenBits();

			// We write the node outdegree
			writeOutdegree( graphObs, outd );

			if ( STATS ) outdegreeStats.println( outd );

			if ( outd > list[ currIndex ].length ) list[ currIndex ] = IntArrays.ensureCapacity( list[ currIndex ], outd );

			// The successor list we are going to compress and write out
			System.arraycopy( nodeIterator.successorArray(), 0, list[ currIndex ], 0, outd );
			listLen[ currIndex ] = outd;

			if ( outd > 0 ) {
			
				// Now we check the best candidate for compression.
				best = Integer.MAX_VALUE;
				bestIndex = -1;

				refCount[ currIndex ] = -1;

				for( j = 0; j < cyclicBufferSize; j++ ) {
					cand = ( currNode - j + cyclicBufferSize ) % cyclicBufferSize;
					if ( refCount[ cand ] < maxRefCount && listLen[ cand ] != 0
						 && ( t = diffComp( bitCount, currNode, j, list[ cand ], listLen[ cand ], list[ currIndex ], listLen[ currIndex ], false ) ) < best ) {
						best = t;
						bestIndex = cand;
					}
				}

				if ( ASSERTS ) assert bestIndex >= 0;
				refCount[ currIndex ] = refCount[ bestIndex ] + 1;
				diffComp( graphObs, currNode, ( currNode - bestIndex + cyclicBufferSize ) % cyclicBufferSize, list[ bestIndex ], 
					listLen[ bestIndex ], list[ currIndex ], listLen[ currIndex], true );
				
				totLinks += outd;
				totRef += refCount[ currIndex ];
				totDist += ( currNode - bestIndex + cyclicBufferSize ) % cyclicBufferSize;
			}

				
			if ( ( currNode + 1 ) % 1000000 == 0 ) System.err.print( "[" +
																	"bits/link=" + format( (double)graphObs.writtenBits() / ( totLinks != 0 ? totLinks : 1 ) ) +
																	", bits/node=" + format( (double)graphObs.writtenBits() / ( currNode ) ) +
																	", avgref=" + format( ( double )totRef / currNode ) +
																	", avgdist=" + format( ( double )totDist / currNode ) +
																	"]" );

			if ( pm != null ) pm.update();
		}
		
		// We write the final offset to the offset stream.
		writeOffset( offsetObs, (int)( graphObs.writtenBits() - bitOffset ) );

		graphObs.close();
		offsetObs.close();

		if ( pm != null ) {
			pm.stop();
			System.err.println( " done." );
			System.err.println( pm );
		}

		// Finally, we save all data related to this graph in a property file.
		final Properties properties = new Properties();
		properties.setProperty( "basename", basename.toString() );
		properties.setProperty( "nodes", String.valueOf( n ) );
		properties.setProperty( "arcs", String.valueOf( totLinks ) );
		properties.setProperty( "windowsize", String.valueOf( windowSize ) );
		properties.setProperty( "maxrefcount", String.valueOf( maxRefCount ) );
		properties.setProperty( "minintervallength", String.valueOf( minIntervalLength ) );
		if ( residualCoding == ZETA ) properties.setProperty( "zetak", String.valueOf( zetaK ) );
		properties.setProperty( "compressionflags", flags2String( flags ).toString() );
		properties.setProperty( "avgref", format( ( double ) totRef / n ) );
		properties.setProperty( "avgdist", format( ( double ) totDist / n ) );
		properties.setProperty( "bitsperlink", format( ( double )graphObs.writtenBits() / totLinks ) );
		properties.setProperty( "bitspernode", format( ( double )graphObs.writtenBits() / n ) );
		properties.setProperty( "graphclass", this.getClass().toString() );
		properties.setProperty( "version", String.valueOf( BVGRAPH_VERSION ) );
		final FileOutputStream propertyFile = new FileOutputStream( basename + ".properties" );
		properties.store( propertyFile, "BVGraph properties" );

		propertyFile.close();

		if ( STATS ) {
			offsetStats.close();
			referenceStats.close();
			outdegreeStats.close();
			blockCountStats.close();
			blockStats.close();
			intervalCountStats.close();
			leftStats.close();
			lenStats.close();
			residualCountStats.close();
			residualStats.close();
		}
	}

	
	/** Write the offset file to a given bit stream.
	 * @param obs the output bit stream to which offsets will be written.
	 * @param pm a progress meter, or <code>null</code>.  
	 */
	public void writeOffsets( final OutputBitStream obs, final ProgressMeter pm ) throws IOException {
		final BVGraphNodeIterator nodeIterator = (BVGraphNodeIterator) nodeIterator( 0 );
		int n = numNodes();
		long lastOffset = 0;
		while( n-- != 0 ) {
				// We fetch the current position of the underlying input bit stream, which is at the start of the next node.
				writeOffset( obs, (int)( nodeIterator.ibs.readBits() - lastOffset ) );
				lastOffset = nodeIterator.ibs.readBits();
				nodeIterator.nextInt();
				nodeIterator.outdegree();
				nodeIterator.successorArray();
				if ( pm != null ) pm.update();
		}
		writeOffset( obs, (int)( nodeIterator.ibs.readBits() - lastOffset ) );
	}
		

	/** Reads an immutable graph and stores it as a {@link BVGraph}.
	 */

	public static void main( String args[] ) throws ClassNotFoundException, IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, IOException {
		String source = null, dest = null;
		Class graphClass = null;
		int windowSize = -1, maxRefCount = -1, minIntervalLength = -1, zetaK = - 1, flags = 0;
		boolean offline = false, writeOffsets = false;
		int quantum = 100000;

		int ch;
		LongOpt longopts[] = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "comp", LongOpt.REQUIRED_ARGUMENT, null, 'c' ),
			new LongOpt( "window-size", LongOpt.REQUIRED_ARGUMENT, null, 'w' ),
			new LongOpt( "max-ref-count", LongOpt.REQUIRED_ARGUMENT, null, 'm' ),
			new LongOpt( "graph-class", LongOpt.REQUIRED_ARGUMENT, null, 'g' ),
			new LongOpt( "min-interval-length", LongOpt.REQUIRED_ARGUMENT, null, 'i' ),
			new LongOpt( "zeta-k", LongOpt.REQUIRED_ARGUMENT, null, 'k' ),
			new LongOpt( "offline", LongOpt.NO_ARGUMENT, null, 'o' ),
			new LongOpt( "offsets", LongOpt.NO_ARGUMENT, null, 'O' ),
		};

		Getopt g = new Getopt( "BVGraph", args, "Ooc:w:m:k:g:i:q:h", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: BVGraph [OPTIONS] SOURCE [DEST]" );
			System.err.println( "Compresses differentially a graph. The SOURCE and DEST parameters are" );
			System.err.println( "basenames from which suitable file names will be stemmed." );
			System.err.println( "" );
			System.err.println( "If DEST is omitted, no recompression is performed. This is useful" );
			System.err.println( "in conjunction with --offsets." );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -c, --comp                 a compression flag (may be specified several times)" );
			System.err.println( "  -w, --window-size          reference window size (default: " + DEFAULT_WINDOW_SIZE + ", 0 to disable)" );
			System.err.println( "  -m, --max-ref-count        maximum number of backward references (default: " + DEFAULT_MAX_REF_COUNT + ")" );
			System.err.println( "  -i, --min-interval-length  minimum length of an interval (default: " + DEFAULT_MIN_INTERVAL_LENGTH + ", 0 to disable)" );
			System.err.println( "  -k, --zeta-k               the k parameter for zeta-k codes (default: " + DEFAULT_ZETA_K + ")" );
			System.err.println( "  -g, --graph-class          forces a Java class to load the source graph" );
			System.err.println( "  -o, --offline              use the offline load method to reduce memory consumption" );
			System.err.println( "  -O, --offsets              generates offsets for the source graph" );
			System.err.println( "  -q, --quantum              the progress meter quantum (default: " + quantum + ")" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help          print this help screen" );
			System.err.println( "" );
			return;

		case 'c':
			try {
				flags |= BVGraph.class.getField( g.getOptarg() ).getInt( BVGraph.class );
			}
			catch ( Exception notFound ) {
				System.err.println( "Compression method " + g.getOptarg() + " unknown." );
				return;
			}
			break;

		case '?':
			return;

		case 'w':
			windowSize = Integer.parseInt( g.getOptarg() );
			break;

		case 'k':
			zetaK = Integer.parseInt( g.getOptarg() );
			break;

		case 'm':
			maxRefCount = Integer.parseInt( g.getOptarg() );
			break;

		case 'i':
			minIntervalLength = Integer.parseInt( g.getOptarg() );
			break;

		case 'g':
			graphClass = Class.forName( g.getOptarg(), true, ClassLoader.getSystemClassLoader() );
			break;

		case 'o':
			offline = true;
			break;

		case 'O':
			writeOffsets = true;
			break;

		case 'q':
			quantum = Integer.parseInt( g.getOptarg() );
			break;

		default:
			break;
		}

		if ( args.length - g.getOptind() != 1 && args.length - g.getOptind() != 2 ) {
			System.err.println( "Wrong number (" + ( args.length - g.getOptind() ) + ") of arguments." );
			return;
		}
		else {
			source = args[ g.getOptind() ];
			if ( args.length - g.getOptind() == 2 ) dest = args[ g.getOptind() + 1 ];
		}

		final ProgressMeter pm = new ProgressMeter();
		final ImmutableGraph graph;
		
		pm.quantum( quantum );

		if ( graphClass != null ) {
			if ( offline ) graph = (ImmutableGraph)graphClass.getMethod( "loadOffline", new Class[] { CharSequence.class } ).invoke( graphClass, new Object[] { source } );
			else graph = (ImmutableGraph)graphClass.getMethod( "loadSequential", new Class[] { CharSequence.class, ProgressMeter.class } ).invoke( graphClass, new Object[] { source, pm } );
		}
		else graph = offline ? ImmutableGraph.loadOffline( source, pm ) : ImmutableGraph.loadSequential( source, pm );

		if ( dest != null )	BVGraph.store( graph, dest, windowSize, maxRefCount, minIntervalLength, zetaK, flags, pm );
		else {
			if ( writeOffsets ) {
				if ( graph instanceof BVGraph ) {
					final OutputBitStream offsets = new OutputBitStream( graph.basename() + ".offsets", 64 * 1024 );
					pm.expectedUpdates( graph.numNodes() );
					pm.start( "Writing offsets..." );
					((BVGraph)graph).writeOffsets( offsets, pm );
					offsets.close();
					pm.done();
				}
				else System.err.println( "The sources graph is not a BVGraph" );
			}
		}
	}
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
