package it.unimi.dsi.webgraph;

/*		 
 * Copyright (C) 2003, 2004, 2005 Paolo Boldi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntIterators;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

/** A simple abstract class representing an immutable graph.
 *
 * <P>Subclasses of this class are used to create and access <em>immutable graphs</em>, that is,
 * graphs that are computed once for all, stored conveniently, and then accessed repeatedly.
 * Moreover, immutable graphs are usually very large&mdash;so large that two such graphs may not
 * fit into central memory (the main example being a sizable portion of the web).
 *
 * <P>A subclass of this class must implement methods to obtain the {@linkplain
 * #numNodes() number of nodes}, the {@linkplain #outdegree(int) outdegree of a
 * node} and the successors of a node (either {@link #successors(int)}
 * or {@link #successorArray(int)}). Additionally, it may provide methods to
 * obtain the {@linkplain #numNodes() number of arcs}, and a {@linkplain #basename() basename}.
 *
 * <P>This class provides {@link #equals(Object)} and {@link #hashCode()} methods that consider
 * two graph equals if they have the same size and all their successor lists are equal.
 *  
 * <H2>Building an immutable graph</H2> 
 *
 * <P><strong>Warning</strong>: the serialisation scheme has been slightly modified
 * as of version 1.1. Now all static load methods must return exactly an {@link ImmutableGraph}
 * (otherwise they cannot be overriden by subclasses). This implies that if you want to
 * assign the result of a load method to a variable with a specific type, you must cast it manually.
 * Specific implementations of {@link ImmutableGraph} may also decide to expose internal load methods
 * to make it easier to write load methods for subclasses 
 * (see, e.g., {@link it.unimi.dsi.webgraph.BVGraph#loadInternal(CharSequence, int, ProgressMeter) loadInternal()}).
 *
 * <P>Due to their large size, immutable
 * graphs have a peculiar serialisation scheme. Every subclass of this class
 * <strong>must</strong> implement a number of static methods that create an immutable
 * graph, given a string (usually a basename for a set of files) and, optionally, a {@link it.unimi.dsi.mg4j.util.ProgressMeter}.
 * The signatures that must be implemented are 
 * <UL>
 * <LI><code>ImmutableGraph load(CharSequence, ProgressMeter)</code>;
 * <LI><code>ImmutableGraph load(CharSequence)</code>;
 * <LI><code>ImmutableGraph loadSequential(CharSequence, ProgressMeter)</code>;
 * <LI><code>ImmutableGraph loadSequential(CharSequence)</code>;
 * <LI><code>ImmutableGraph loadOffline(CharSequence, ProgressMeter)</code>;
 * <LI><code>ImmutableGraph loadOffline(CharSequence)</code>.
 * </UL>
 * 
 * <P>The special semantics associated with <code>loadSequential()</code> is that the graph should be loaded
 * in a particular compact form that just allows sequential access. The special semantics associated to <code>loadOffline</code> 
 * is that the immutable graph should be set up, and possibly some metadata could be read from disk, but no
 * actual data is loaded into memory; the class should guarantee that offline sequential access (i.e., by means
 * of {@link #nodeIterator(int)}) is still possible. In other words, in most cases {@link #nodeIterator(int)} will have to be
 * overridden by the subclasses to behave properly even in an offline setting (see {@link #nodeIterator()}).
 * 
 * <P>Note that a simple class may just implement all special forms of graph loading delegating to the standard
 * load method (see, e.g., {@link it.unimi.dsi.webgraph.ASCIIGraph}).
 * 
 * <P>Analogously, a subclass of this class <strong>may</strong> also implement
 * <UL>
 * <LI><code>store(ImmutableGraph, CharSequence, ProgressMeter)</code>;
 * <LI><code>store(ImmutableGraph, CharSequence)</code>.
 * </UL> 
 * 
 * This method must store in compressed form a given immutable graph, using the default values
 * for compression parameters, etc. It is likely, however, that more 
 * of <code>store</code> methods are available, as parameters vary wildly
 * from subclass to subclass. The method {@link #store(Class, ImmutableGraph, CharSequence, ProgressMeter)}
 * invokes by reflection the methods above on the provided class.
 *
 * <P>The standard method to build a new immutable graph is creating a (possibly anonymous) class
 * that extends this class, and save it using a concrete subclass (e.g., {@link it.unimi.dsi.webgraph.BVGraph}). See
 * the source of {@link it.unimi.dsi.webgraph.Transform} for several examples.
 * 
 * <H2>Properties Conventions</H2>
 * 
 * <P>To provide a simple way to load an immutable graph without knowing in advance its class,
 * the following convention may be followed: a graph with basename <var><samp>name</samp></var> may feature
 * a Java property file <samp><var>name</var>.properties</samp> with a property <samp>graphclass</samp>
 * containing the actual class of the graph. In this case, you can use the implementation of the load/store
 * methods contained in this class, similarly to the standard Java serialisation scheme. {@link BVGraph}, for instance,
 * follows this convention, but {@link ASCIIGraph} does not.
 * 
 * <P>The reason why this convention is not enforced is that it is sometimes useful to write lightweight classes,
 * mostly for debugging purposes, whose graph representation is entirely contained in a single file (e.g., {@link ASCIIGraph}).
 * 
 * <H2>Facilities for loading an immutable graph</H2>
 * 
 * <P>{@link ImmutableGraph} provides ready-made implementations of the load methods that work as follows: it
 * opens a property file with the given basename, and looks for the <samp>graphclass</samp> property; then, it simply
 * delegates the actual load to the specified graph class.
 */


public abstract class ImmutableGraph {

	/** Returns the number of nodes of this graph. 
	 *
	 * @return the number of nodes.
	 */
	public abstract int numNodes();

	/** Returns the number of arcs of this graph (optional operation). 
	 *
	 * @return the number of arcs.
	 */
	public long numArcs() {
		throw new UnsupportedOperationException();
	}

	/** Returns a symbolic basename for this graph (optional operation).
	 *
	 * <P>Implementors of this class may provide a basename (usually
	 * a pathname from which various files storing the graph are stemmed).
	 * This method is optional because it is sometimes unmeaningful (e.g.,
	 * for one-off anonymous classes).
	 *
	 * @return the basename.
	 */
	public CharSequence basename() {
		throw new UnsupportedOperationException();
	}

	/** Returns an iterator over the successors of a given node.
	 * 
	 * <P>This implementation just wraps the array returned by {@link #successorArray(int)}. Subclasses 
	 * are encouraged to override this implementation.
	 *
	 * 	 * @param x a node.
	 * @return an iterator over the successors of the node.
	 */
	public IntIterator successors( int x ) {
		return IntIterators.wrap( successorArray( x ), 0, outdegree( x ) );
	}

	/** Returns a reference to an array containing the successors of a given node.
	 * 
	 * <P>The returned array may contain more entries than the outdegree of <code>x</code>.
	 * However, only those with indices from 0 (inclusive) to the outdegree of <code>x</code> (exclusive)
	 * contain valid data.
	 * 
	 * <P>This implementation just unwraps the iterator returned by {@link #successors(int)}. Subclasses 
	 * are encouraged to override this implementation.
	 * 
	 * @param x a node.
	 * @return an array whose first elements are the successors of the node; the array must not
	 * be modified by the caller.
	 */
	public int[] successorArray( int x ) {
		return IntIterators.unwrap( successors( x ) );
	}

	/** Returns the outdegree of a node.
	 * 
	 * @param x a node.
	 * @throws IllegalStateException if called without offsets.
	 * @return the outdegree of the given node.
	 */
	public abstract int outdegree( int x );

	/** Returns a node iterator for scanning the graph sequentially, starting from the given node.
	 *
	 *  <P>This implementation just calls the random-access methods ({@link #successors(int)} and
	 *  {@link #outdegree(int)}). More specific implementations may choose to maintain some extra state
	 *  to make the enumeration more efficient.
	 *
	 *  @param from the node from which the iterator will iterate.
	 *  @return a {@link NodeIterator} for accessing nodes and successors sequentially.
	 */
	public NodeIterator nodeIterator( final int from ) {
		return new NodeIterator() {
				int curr = from - 1;
				final int n = numNodes();

				public int nextInt() {
					if ( ! hasNext() ) throw new java.util.NoSuchElementException();
					return ++curr;
				}

				public boolean hasNext() {
					return ( curr < n - 1 );
				}

				public IntIterator successors() {
					if ( curr == from - 1 ) throw new IllegalStateException();
					return ImmutableGraph.this.successors( curr );
				}

				public int outdegree() {
					if ( curr == from - 1 ) throw new IllegalStateException();
					return ImmutableGraph.this.outdegree( curr );
				}
				
			};
	}
	
	/** Returns a node iterator for scanning the graph sequentially, starting from the first node.
	 *
	 *  @return a {@link NodeIterator} for accessing nodes and successors sequentially.
	 */
	public NodeIterator nodeIterator() {
		return nodeIterator( 0 );
	}

	public String toString() {
		MutableString ms = new MutableString();
		IntIterator ii;

		ms.append( "Nodes: " + numNodes() + "\nArcs: " + numArcs() + "\n" );
		for ( int i = 0; i < numNodes(); i++ ) {
			ms.append( "Successors of " + i + " (degree " + outdegree( i ) + "):" );
			ii = successors( i );
			while ( ii.hasNext() ) 
				ms.append( " " + ii.nextInt() );
			ms.append( "\n" );
		}
		return ms.toString();
	}


	/** Creates a new {@link ImmutableGraph} by loading offline a graph file.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param basename the basename of the graph.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */


	public static ImmutableGraph loadSequential( CharSequence basename ) throws IOException {
		return load( "loadSequential", basename );
	}

	/** Creates a new {@link ImmutableGraph} by loading offline a graph file, using a progress meter.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param basename the basename of the graph.
	 * @param pm a progress meter used while loading the graph, or <code>null</code>.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */

	public static ImmutableGraph loadSequential( CharSequence basename, ProgressMeter pm ) throws IOException {
		return load( "loadSequential", basename, pm );
	}

	/** Creates a new {@link ImmutableGraph} by loading a graph file from disk to memory, without 
	 *  offsets.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param basename the basename of the graph.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */

	public static ImmutableGraph loadOffline( CharSequence basename ) throws IOException {
		return load( "loadOffline", basename );
	}


	/** Creates a new {@link ImmutableGraph} by loading a graph file from disk to memory, without 
	 *  offsets, using a progress meter.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param basename the basename of the graph.
	 * @param pm a progress meter used while loading the graph, or <code>null</code>.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */

	public static ImmutableGraph loadOffline( CharSequence basename, ProgressMeter pm ) throws IOException {
		return load( "loadOffline", basename, pm );
	}


	/** Creates a new {@link ImmutableGraph} by loading a graph file from disk to memory, with 
	 *  all offsets.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param basename the basename of the graph.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */


	public static ImmutableGraph load( CharSequence basename ) throws IOException {
		return load( "load", basename );
	}

	/** Creates a new {@link ImmutableGraph} by loading a graph file from disk to memory, with 
	 *  all offsets, using a progress meter.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param basename the basename of the graph.
	 * @param pm a progress meter used while loading the graph, or <code>null</code>.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */

	public static ImmutableGraph load( CharSequence basename, ProgressMeter pm) throws IOException {
		return load( "load", basename, pm );
	}

	private static final ProgressMeter UNUSED = new ProgressMeter();

	/** Creates a new {@link ImmutableGraph} by loading a graph file from disk to memory, with 
	 *  all offsets, using no progress meter.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param basename the basename of the graph.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */	
	private static ImmutableGraph load( CharSequence method, CharSequence basename ) throws IOException {
		return load( method, basename, UNUSED );
	}

	/** Creates a new immutable graph by loading a graph file from disk to memory, delegating the 
	 *  actual loading to the class specified in the <samp>graphclass</samp> property within the property
	 *  file (named <samp><var>basename</var>.properties</samp>). The exact load method to be used 
	 *  depends on the <code>method</code> argument.
	 * 
	 * <P>This method uses the properties convention described in the {@linkplain ImmutableGraph introduction}.
	 *
	 * @param method the name of the method to be used to load the graph (<samp>load</samp>, <samp>loadSequential</samp>, <samp>loadOffline</samp>).
	 * @param basename the basename of the graph.
	 * @param pm the progress meter; it can be <code>null</code>.
	 * @return an {@link ImmutableGraph} containing the specified graph.
	 * @throws IOException if an I/O exception occurs while reading the graph.
	 */	
	protected static ImmutableGraph load( CharSequence method, CharSequence basename, ProgressMeter pm ) throws IOException {
		final FileInputStream propertyFile = new FileInputStream( basename + ".properties" );
		final Properties properties = new Properties();
		String graphClassName;
		properties.load( propertyFile );
		propertyFile.close();

		if ( ( graphClassName = properties.getProperty( "graphclass" ) ) == null ) throw new IOException( "The property file for " + basename + " does not contain a graphclass property" );
		// Small kluge to fix old usage of toString() instead of getName();
		if ( graphClassName.startsWith( "class " ) ) graphClassName = graphClassName.substring( 6 );
		final Class graphClass;
		
		ImmutableGraph graph = null;

		try {
			graphClass = Class.forName( graphClassName );

			if ( pm == UNUSED ) graph = (ImmutableGraph)graphClass.getMethod( method.toString(), new Class[] { CharSequence.class } ).invoke( graphClass, new Object[] { basename } );
			else graph = (ImmutableGraph)graphClass.getMethod( method.toString(), new Class[] { CharSequence.class, ProgressMeter.class } ).invoke( graphClass, new Object[] { basename, pm } );
		} catch ( InvocationTargetException e ) {
			if ( e.getCause() instanceof IOException ) throw (IOException) e.getCause();
			else throw new RuntimeException( e );
		} catch( Exception e ) {
			throw new RuntimeException( e );
		}

		return graph;
	}




	/** Stores an immutable graph using a specified subclass and a progress meter.
	 * 
	 * <P>This method is a useful shorthand that invoke by reflection the store method of a given subclass.
	 * Note, however, that usually a subclass will provide more refined store methods with more parameters.
	 *
	 * @param graphClass the subclass of {@link ImmutableGraph} that should store the graph.
	 * @param graph the graph to store.
	 * @param basename the basename.
	 * @param pm a progress meter, or <code>null</code>.
	 */

	public static void store( final Class graphClass, final ImmutableGraph graph, final CharSequence basename, final ProgressMeter pm ) throws IOException {
		if ( ! ImmutableGraph.class.isAssignableFrom( graphClass ) ) throw new ClassCastException( graphClass.getName() + " is not a subclass of ImmutableGraph" );
		try {
			if ( pm == UNUSED ) graphClass.getMethod( "store", new Class[] { ImmutableGraph.class, CharSequence.class } ).invoke( graphClass, new Object[] { graph, basename } );
			else graphClass.getMethod( "store", new Class[] { ImmutableGraph.class, CharSequence.class, ProgressMeter.class } ).invoke( graphClass, new Object[] { graph, basename, pm } );
		} catch ( InvocationTargetException e ) {
			if ( e.getCause() instanceof IOException ) throw (IOException) e.getCause();
			else throw new RuntimeException( e );
		} catch( Exception e ) {
			throw new RuntimeException( e );
		}
	}
	
	/** Stores an immutable graph using a specified subclass.
	 * 
	 * @param graphClass the subclass of {@link ImmutableGraph} that should store the graph.
	 * @param graph the graph to store.
	 * @param basename the basename.
	 * @see #store(Class, ImmutableGraph, CharSequence, ProgressMeter)
	 */

	public static void store( final Class graphClass, final ImmutableGraph graph, final CharSequence basename ) throws IOException {
		store( graphClass, graph, basename, UNUSED );
	}

	/** Compare this immutable graph to another object.
	 *
	 * @return true iff the given object is an immutable graph of the same size, and
	 * the successor list of every node of this graph is equal to the successor list of the corresponding node of <code>o</code>.
	 */

	public boolean equals( final Object o ) {
		if ( ! (o instanceof ImmutableGraph) ) return false;
		final ImmutableGraph g = (ImmutableGraph) o;
		int n = numNodes();
		if ( n != g.numNodes() ) return false;
		final NodeIterator i = nodeIterator(), j = g.nodeIterator();
		IntIterator s, t;
		int d;
		// TODO: replace this with an equals() method for iterators and / or
		// an equals() method for arrays that takes offset and length (fastutil).
		while( n-- != 0 ) { 
			i.nextInt();
			j.nextInt();
			if ( ( d = i.outdegree() ) != j.outdegree() ) return false;
			s = i.successors();
			t = j.successors();
			while( d-- != 0 ) if ( s.nextInt() != t.nextInt() ) return false;
		}
			
		return true;
	}
	
	/** Returns a hash code for this immutable graph.
	 *
	 * @return a hash code for this immutable graph.
	 */

	public int hashCode() {
		int n = numNodes(), h = -1;
		final NodeIterator i = nodeIterator();
		IntIterator s;
		int d;
		while( n-- != 0 ) { 
			h = h * 31 + i.nextInt();
			s = i.successors();
			d = i.outdegree();
			while( d-- != 0 ) h = h * 31 + s.nextInt();
		}

		return h;
	}
	
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
